from euphorie.content.passwordpolicy import EuphoriePasswordPolicy
from euphorie.content.utils import REGION_NAMES
from plone import api
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.dexterity.utils import createContentInContainer
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import _createObjectByType
from Products.PlonePAS.plugins.passwordpolicy import PasswordPolicyPlugin
from Products.PluggableAuthService.interfaces.plugins import IValidationPlugin
from zope.interface import alsoProvides

import logging


log = logging.getLogger(__name__)


def setupVarious(context):
    site = api.portal.get()
    disableRedirectTracking(site)
    setupInitialContent(site)
    setupVersioning(site)
    registerPasswordPolicy(site)
    setupSecureSessionCookie(site)


COUNTRIES = {
    "at": ("Austria", "eu-member"),
    "be": ("Belgium", "eu-member"),
    "bg": ("Bulgaria", "eu-member"),
    "se": ("Sweden", "eu-member"),
    "cy": ("Cyprus", "eu-member"),
    "cz": ("The Czech Republic", "eu-member"),
    "de": ("Germany", "eu-member"),
    "dk": ("Denmark", "eu-member"),
    "ee": ("Estonia", "eu-member"),
    "es": ("Spain", "eu-member"),
    "fi": ("Finland", "eu-member"),
    "fr": ("France", "eu-member"),
    "gb": ("The United Kingdom", "eu-member"),
    "gr": ("Greece", "eu-member"),
    "hu": ("Hungary", "eu-member"),
    "ie": ("Ireland", "eu-member"),
    "it": ("Italy", "eu-member"),
    "lt": ("Lithuania", "eu-member"),
    "lu": ("Luxembourg", "eu-member"),
    "lv": ("Latvia", "eu-member"),
    "mt": ("Malta", "eu-member"),
    "nl": ("The Netherlands", "eu-member"),
    "pl": ("Poland", "eu-member"),
    "pt": ("Portugal", "eu-member"),
    "ro": ("Romania", "eu-member"),
    "sk": ("Slovakia", "eu-member"),
    "si": ("Slovenia", "eu-member"),
    "li": ("Liechtenstein", "efta"),
    "no": ("Norway", "efta"),
    "ch": ("Switzerland", "efta"),
    "hr": ("Republic of Croatia", "candidate-eu"),
    "is": ("Republic of Iceland", "candidate-eu"),
    "mk": ("F.Y.R. Macedonia", "candidate-eu"),
    "me": ("Montenegro", "candidate-eu"),
    "tr": ("Republic of Turkey", "candidate-eu"),
    "al": ("Republic of Albania", "potential-candidate-eu"),
    "ba": ("Bosnia and Herzegovina", "potential-candidate-eu"),
    "cs": ("Kosovo", "potential-candidate-eu"),
    "rs": ("Republic of Serbia", "potential-candidate-eu"),
}

for i in REGION_NAMES.items():
    COUNTRIES[i[0]] = (i[1], "region")


def setupInitialContent(site):
    present = site.objectIds()
    for obj in ["Members", "events", "news"]:
        if obj in present:
            site.manage_delObjects([obj])
            log.info("Removed default Plone %s folder", obj)

    if "sectors" not in present:
        site.invokeFactory(
            "euphorie.sectorcontainer",
            "sectors",
            title="Surveys",
        )
        mt = getToolByName(site, "portal_membership")
        mt.setMembersFolderById("sectors")
        log.info("Added sectors folder")

    sectors = site.sectors
    for country_id, info in COUNTRIES.items():
        (title, country_type) = info
        if country_id not in sectors:
            sectors.invokeFactory(
                "euphorie.country",
                country_id,
                title=title,
                country_type=country_type,
            )
            log.info("Added country %s (%s)", country_id, title)
        country = sectors[country_id]
        if "help" not in country:
            createContentInContainer(
                country,
                "euphorie.page",
                id="help",
                title="Help",
                checkConstraints=False,
            )
            log.info("Added help section for country %s (%s)", country_id, title)
        help = country["help"]
        if not INavigationRoot.providedBy(help):
            alsoProvides(help, INavigationRoot)
            log.info(
                "Made help for country %s (%s) a navigation root.",
                country_id,
                title,
            )

    if "client" not in present:
        site.invokeFactory("euphorie.client", "client", title="Client")
        api.content.transition(site.client, to_state="published")
        log.info("Added Euphorie client instance")

    if "documents" not in present:
        site.invokeFactory("euphorie.folder", "documents", title="Documents")
        log.info("Added documents folder")
    documents = site.documents

    if not INavigationRoot.providedBy(documents):
        alsoProvides(documents, INavigationRoot)
        log.info("Made documentation folder a navigation root.")

    lt = getToolByName(site, "portal_languages")
    present_languages = documents.objectIds()
    for code, name in lt.listSupportedLanguages():
        if code not in present_languages:
            documents.invokeFactory(
                "euphorie.documentation",
                code,
                title=name,
            )
            log.info("Added documentation folder for %s (%s)", name, code)
        docs = documents[code]
        if "help" not in docs:
            createContentInContainer(
                docs,
                "euphorie.help",
                id="help",
                checkConstraints=False,
            )
            log.info("Added online help text for language %s (%s)", name, code)
        if "appendix" not in docs:
            _createObjectByType(
                "euphorie.page",
                docs,
                "appendix",
                title="Appendix",
            )
            log.info("Added appendix folder for language %s (%s)", name, code)


def disableRedirectTracking(site):
    # Add additional setup code here
    from plone.app.redirector.interfaces import IRedirectionStorage
    from zope.component import getSiteManager
    from zope.interface.interfaces import IComponentRegistry

    sm = getSiteManager(site)
    if sm is None or not IComponentRegistry.providedBy(sm):
        log.warning(
            "Failed to find a site manager, can not remove "
            "IRedirectionStorage utility"
        )
        return

    sm.unregisterUtility(provided=IRedirectionStorage)


def setupVersioning(site):
    repository = site.portal_repository
    if "euphorie.survey" not in repository.getVersionableContentTypes():
        repository.setVersionableContentTypes(["euphorie.survey"])
        log.info("Enabled versioning for survey versions.")


def registerPasswordPolicy(site):
    pas = api.portal.get_tool("acl_users")

    # Deactivate the default policy
    for oid in pas.objectIds([PasswordPolicyPlugin.meta_type]):
        if oid in pas.plugins._getPlugins(IValidationPlugin):
            pas.plugins.deactivatePlugin(
                IValidationPlugin,
                oid,
            )

    # Activate the Euphorie policy
    if not pas.objectIds([EuphoriePasswordPolicy.meta_type]):
        plugin = EuphoriePasswordPolicy(
            EuphoriePasswordPolicy.id,
            EuphoriePasswordPolicy.meta_type,
        )
        pas._setObject(plugin.getId(), plugin)
        plugin = getattr(pas, plugin.getId())

        infos = [
            info
            for info in pas.plugins.listPluginTypeInfo()
            if plugin.testImplements(info["interface"])
        ]
        plugin.manage_activateInterfaces([info["id"] for info in infos])
        for info in infos:
            for i in range(len(pas.plugins.listPluginIds(info["interface"]))):
                pas.plugins.movePluginsUp(
                    info["interface"],
                    [plugin.getId()],
                )


def setupSecureSessionCookie(site):
    session = api.portal.get_tool("acl_users").get("session")
    if not session:
        return
    if not session.secure:
        session.secure = True



from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.content import MessageFactory as _
from plone import api
from Products.Five import BrowserView
from zExceptions import Unauthorized
from zope.component import getMultiAdapter


class Lock(BrowserView):
    """Lock or unlock a User account.

    View name: @@lock
    """

    def __call__(self):
        if self.request.method != "POST":
            raise Unauthorized
        authenticator = getMultiAdapter(
            (self.context, self.request), name="authenticator"
        )
        if not authenticator.verify():
            raise Unauthorized

        self.context.locked = locked = self.request.form.get("action", "lock") == "lock"
        if locked:
            api.portal.show_message(
                _(
                    "message_user_locked",
                    default='Account "${title}" has been locked.',
                    mapping=dict(title=self.context.title),
                ),
                self.request,
                "success",
            )
        else:
            api.portal.show_message(
                _(
                    "message_user_unlocked",
                    default='Account "${title}" has been unlocked.',
                    mapping=dict(title=self.context.title),
                ),
                self.request,
                "success",
            )

        country = aq_parent(aq_inner(self.context))
        self.request.response.redirect("%s/@@manage-users" % country.absolute_url())



"""
Company
-------

View and update the company survey.
"""

from .. import MessageFactory as _
from datetime import datetime
from euphorie.client import model
from plone import api
from plone.autoform import directives
from plone.autoform.form import AutoExtensibleForm
from plone.memoize.view import memoize
from plone.supermodel.model import Schema
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form import form
from z3c.form.form import applyChanges
from zope import schema
from zope.deprecation import deprecate
from zope.interface import directlyProvides
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class CompanySchema(Schema):
    # Note the ideal type, but there is no convenient country field
    country = schema.TextLine(
        title=_("label_company_country", default="Your country"), required=False
    )

    employees = schema.Choice(
        title=_("label_employee_numbers", default="Number of employees"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(
                    "1-9",
                    title=_("employee_numbers_1_to_9", default="1 to 9 employees"),
                ),
                SimpleTerm(
                    "10-49",
                    title=_("employee_numbers_10_to_49", default="10 to 49 employees"),
                ),
                SimpleTerm(
                    "50-249",
                    title=_(
                        "employee_numbers_50_to_249", default="50 to 249 employees"
                    ),
                ),
                SimpleTerm(
                    "250+",
                    title=_(
                        "employee_numbers_250_or_more", default="250 or more employees"
                    ),
                ),
            ]
        ),
        required=False,
    )

    conductor = schema.Choice(
        title=_("label_conductor", default="The risk assessment was conducted by"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("staff", title=_("conductor_staff", default="own staff")),
                SimpleTerm(
                    "third-party",
                    title=_(
                        "conductor_third_party",
                        default="an external consultant or service provider",
                    ),
                ),
                SimpleTerm(
                    "both",
                    title=_(
                        "conductor_both",
                        default="both own staff and an external consultant or "
                        "service provider",
                    ),
                ),
            ]
        ),
        required=False,
    )

    referer = schema.Choice(
        title=_(
            "label_referer",
            default="Through which channel did you learn about this " "tool?",
        ),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(
                    "employers-organisation",
                    title=_(
                        "referer_employers_organisation",
                        default="an employers' organisation",
                    ),
                ),
                SimpleTerm(
                    "trade-union",
                    title=_(
                        "referer_trade_union", default="a trade union organisation"
                    ),
                ),
                SimpleTerm(
                    "national-public-institution",
                    title=_(
                        "referer_national_public_institution",
                        default="a national public " "institution/administration",
                    ),
                ),
                SimpleTerm(
                    "eu-institution",
                    title=_(
                        "referer_eu_institution",
                        default="an European institution/administration",
                    ),
                ),
                SimpleTerm(
                    "health-safety-experts",
                    title=_(
                        "referer_health_safety_expert",
                        default="health and safety experts",
                    ),
                ),
                SimpleTerm("other", title=_("referer_other", default="other channel")),
            ]
        ),
        required=False,
    )

    workers_participated = schema.Choice(
        title=_(
            "label_workers_participated",
            default="Workers were invited to participate in the " "risk assessment",
        ),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(True, title=_("label_yes", default="Yes")),
                SimpleTerm(False, title=_("label_no", default="No")),
            ]
        ),
        required=False,
    )

    needs_met = schema.Choice(
        title=_("label_needs_met", default="Did this OiRA tool meet your needs?"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(True, title=_("label_yes", default="Yes")),
                SimpleTerm(False, title=_("label_no", default="No")),
            ]
        ),
        required=False,
    )

    recommend_tool = schema.Choice(
        title=_(
            "label_recommend_tool",
            default="Would you recommend this OiRA tool to an enterprise "
            "similar to yours?",
        ),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm(True, title=_("label_yes", default="Yes")),
                SimpleTerm(False, title=_("label_no", default="No")),
            ]
        ),
        required=False,
    )

    directives.mode(timestamp="hidden")
    timestamp = schema.Datetime(
        title="Timestamp",
        required=False,
    )


class Company(AutoExtensibleForm, form.Form):
    """Update the company details.

    View name: @@report_company
    """

    variation_class = "variation-risk-assessment"

    schema = CompanySchema
    template = ViewPageTemplateFile("templates/report_company.pt")
    company_class = model.Company

    @property
    def session(self):
        return self.context.session

    @property
    def default_company_values(self):
        """The values we use to create a new company."""
        return {"session": self.session}

    @property
    @memoize
    def company(self):
        """Get or create the company object for this session."""
        company = (
            model.Session.query(self.company_class)
            .filter(self.company_class.session == self.session)
            .first()
        )
        if not company:
            company = self.company_class(**self.default_company_values)

        # This is used to make the company object the context of the form
        directlyProvides(company, self.schema)
        return company

    @deprecate(
        "Deprecated in version 15.2.1.dev0. "
        "This was a trick to add the company attribute to the instance. "
        "Now company is a property."
    )
    def _assertCompany(self):
        return

    def countries(self):
        names = [
            {"id": key.lower(), "title": value}
            for (key, value) in self.request.locale.displayNames.territories.items()
        ]
        names.sort(key=lambda c: c["title"])
        return names

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    def update(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        super().update()
        self._assertCompany()

    def getContent(self):
        self._assertCompany()
        return self.company

    def applyChanges(self, data):
        content = self.getContent()
        applyChanges(self, content, data)

    @button.buttonAndHandler("Previous")
    def handlePrevious(self, action):
        (data, errors) = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        data["timestamp"] = datetime.now()
        self.applyChanges(data)
        url = "%s/@@report" % self.context.absolute_url()
        self.request.response.redirect(url)

    @button.buttonAndHandler("Next")
    def handleNext(self, action):
        (data, errors) = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        data["timestamp"] = datetime.now()
        self.applyChanges(data)
        url = "%s/@@report_view" % self.context.absolute_url()
        self.request.response.redirect(url)

    @button.buttonAndHandler("Skip")
    def handleSkip(self, action):
        # XXX: This a hack. We need to know if a company report has been
        # skipped but can't add new SQL columns. So we mark the country 'xx'.
        data = {
            "conductor": None,
            "country": "xx",
            "employees": None,
            "referer": None,
            "workers_participated": None,
            "needs_met": None,
            "recommend_tool": None,
        }
        self.applyChanges(data)
        url = "%s/@@report_view" % self.context.absolute_url()
        self.request.response.redirect(url)



from AccessControl import Unauthorized
from datetime import timezone
from euphorie.client import MessageFactory as _
from euphorie.client.browser.base import BaseView
from euphorie.client.model import Account
from euphorie.client.model import Consultancy
from euphorie.client.model import OrganisationMembership
from euphorie.client.model import SessionEvent
from euphorie.client.utils import CreateEmailTo
from json import JSONDecodeError
from plone import api
from plone.memoize import instance
from plone.memoize.view import memoize
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

import html
import json
import logging


logger = logging.getLogger(__name__)


class ConsultancyBaseView(BaseView):
    @property
    def organisation(self):
        return self.context.session.account.organisation

    @property
    def is_admin(self):
        if not self.organisation:
            return False
        organisation_view = api.content.get_view(
            name="organisation",
            context=self.webhelpers.country_obj,
            request=self.request,
        )
        return organisation_view.get_member_role_id(
            self.organisation, self.webhelpers.get_current_account()
        ) in ["admin", "owner"]

    @property
    def consultants(self):
        if not self.organisation:
            return []
        return (
            self.sqlsession.query(Account)
            .join(
                OrganisationMembership,
                OrganisationMembership.member_id == Account.id,
            )
            .filter(OrganisationMembership.owner_id == self.organisation.owner_id)
            .filter(OrganisationMembership.member_role == "consultant")
            .order_by(Account.loginname)
            .all()
        )


# TODO make this an adapter
class SessionValidated:
    def __init__(self, context):
        self.context = context

    @property
    @instance.memoize
    def note(self):
        try:
            return json.loads(self.context.note)
        except JSONDecodeError:
            logger.warning(
                "Could not load note for session_event %r (%r)",
                self.context.id,
                self.context.note,
            )
            return {}

    @property
    def raw_time(self):
        return self.context.time.replace(tzinfo=timezone.utc)

    @property
    def consultant_email(self):
        return self.note["consultant_email"]

    @property
    def consultant_name(self):
        return self.note["consultant_name"]


class ConsultancyView(ConsultancyBaseView):
    """ """

    variation_class = "variation-risk-assessment"

    @property
    @memoize
    def validated_info(self):
        events = (
            self.sqlsession.query(SessionEvent)
            .filter(SessionEvent.session_id == self.context.session.id)
            .filter(SessionEvent.action == "validated")
            .order_by(SessionEvent.time.desc())
        )
        event = events.first()
        return SessionValidated(event)

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        return super().__call__()


class PanelRequestValidation(ConsultancyBaseView):
    """ """

    default_target_view = "@@consultancy"
    email_template = ViewPageTemplateFile("templates/notify-request-validation.pt")

    @property
    @memoize
    def consultant(self):
        consultant_id = self.request.form.get("consultant")
        consultant = (
            self.sqlsession.query(Account).filter(Account.id == consultant_id).one()
        )
        return consultant

    def notify_consultant(self):
        consultant = self.consultant
        requester = self.webhelpers.get_current_account()
        session = self.context
        body = html.unescape(
            self.email_template(
                consultant=self.consultant.first_name or consultant.title,
                requester=requester.title,
                assessment_title=session.title,
                assessment_link=f"{session.absolute_url()}/@@start",
            )
        )
        subject = api.portal.translate(
            _(
                "subject_request_validation",
                default="Risk assessment validation request",
            )
        )
        email_from_name = api.portal.get_registry_record("plone.email_from_name")
        email_from_address = api.portal.get_registry_record("plone.email_from_address")
        mail = CreateEmailTo(
            email_from_name,
            email_from_address,
            consultant.email,
            subject,
            body,
        )

        api.portal.send_email(
            body=mail,
            recipient=self.consultant.email,
            sender=email_from_address,
            subject=subject,
        )
        logger.info("Sent validation request email to %s", self.consultant.email)

    def handle_POST(self):
        """Handle the POST request."""
        consultancy = Consultancy(
            account=self.consultant,
            session=self.context.session,
        )
        self.context.session.consultancy = consultancy
        event = SessionEvent(
            account_id=self.webhelpers.get_current_account().id,
            session_id=self.context.session.id,
            action="validation_requested",
        )
        self.sqlsession.add(event)

        self.notify_consultant()
        self.redirect()

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        if not self.is_admin:
            raise Unauthorized(
                "Only organisation administrators can request validation"
            )
        return super().__call__()


class PanelValidateRiskAssessment(ConsultancyBaseView):
    """ """

    default_target_view = "@@consultancy"
    email_template = ViewPageTemplateFile("templates/notify-assessment-validated.pt")

    @property
    def organisation_admins(self):
        organisation_view = api.content.get_view(
            name="organisation",
            context=self.webhelpers.country_obj,
            request=self.request,
        )
        for account, membership in organisation_view.get_memberships(self.organisation):
            if membership.member_role == "admin":
                yield account
        yield self.sqlsession.query(Account).filter(
            Account.id == self.organisation.owner_id
        ).one()

    def notify_admins(self):
        consultant = self.webhelpers.get_current_account()
        subject = api.portal.translate(
            _(
                "subject_assessment_validated",
                default="Risk assessment validated",
            )
        )
        email_from_name = api.portal.get_registry_record("plone.email_from_name")
        email_from_address = api.portal.get_registry_record("plone.email_from_address")
        for admin in self.organisation_admins:
            body = self.email_template(
                consultant=consultant.title,
                requester=admin.first_name or admin.title,
                assessment_link=f"{self.context.absolute_url()}/@@consultancy",
            )
            mail = CreateEmailTo(
                email_from_name,
                email_from_address,
                admin.email,
                subject,
                body,
            )

            api.portal.send_email(
                body=mail,
                recipient=admin.email,
                sender=email_from_address,
                subject=subject,
            )
        logger.info("Sent validation confirmation email to organisation admins")

    def handle_POST(self):
        """Handle the POST request."""
        if self.request.form.get("approved", False):
            consultant = self.webhelpers.get_current_account()
            if not self.context.session.is_locked:
                # Lock the assessment in addition to validating it to make sure that it
                # goes back to the locked state when invalidating it
                locked_event = SessionEvent(
                    account_id=api.user.get_current().id,
                    session_id=self.context.session.id,
                    action="lock_set",
                )
                self.sqlsession.add(locked_event)
            validated_event = SessionEvent(
                account_id=consultant.id,
                session_id=self.context.session.id,
                action="validated",
                note=json.dumps(
                    {
                        "consultant_id": consultant.id,
                        "consultant_email": consultant.email,
                        "consultant_name": consultant.title,
                    }
                ),
            )
            self.sqlsession.add(validated_event)
            self.context.session.consultancy.status = "validated"
            self.notify_admins()
        self.redirect()

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        consultancy = self.context.session.consultancy
        if not consultancy or (
            self.webhelpers.get_current_account() != consultancy.account
        ):
            raise Unauthorized(
                "Only the consultant assigned to a risk assessment can validate it"
            )
        return super().__call__()


class PanelInvalidateRiskAssessment(ConsultancyBaseView):
    """ """

    default_target_view = "@@consultancy"

    def handle_POST(self):
        """Handle the POST request."""
        invalidated_event = SessionEvent(
            account_id=api.user.get_current().id,
            session_id=self.context.session.id,
            action="invalidated",
        )
        self.sqlsession.add(invalidated_event)
        self.sqlsession.delete(self.context.session.consultancy)
        self.redirect()

    def __call__(self):
        if not self.webhelpers.can_view_session:
            return self.request.response.redirect(self.webhelpers.client_url)
        if not self.is_admin:
            raise Unauthorized(
                "Only organisation administrators can invalidate an assessment"
            )
        return super().__call__()


class Consultants(BrowserView):
    """Show a country/language specific page about finding consultants"""

    variation_class = "variation-risk-assessment"  # to show left-hand navigation

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def content(self):
        help_folder = self.webhelpers.content_country_obj.help
        for lang in self.webhelpers._getLanguages():
            docs = help_folder.get(lang, None)
            if docs is None:
                continue
            consultants = docs.get("consultants", None)
            if consultants is not None:
                return consultants.body
        # fall back to any available language
        for docs in help_folder.objectValues():
            consultants = docs.get("consultants", None)
            if consultants is not None:
                return consultants.body

        return ""



from email.utils import formataddr
from euphorie.client import MessageFactory as _
from euphorie.client import model
from euphorie.client.mails.base import BaseEmail
from logging import getLogger
from plone import api
from plone.memoize.view import memoize
from z3c.saconfig import Session


logger = getLogger(__name__)


class BaseNotificationEmail(BaseEmail):
    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", api.portal.get(), self.request)

    @property
    def account(self):
        return self.context["account"]

    @property
    def sessions(self):
        return self.context["sessions"]

    @property
    def sender(self):
        from_address = api.portal.get_registry_record(
            "euphorie.notification__email_from_address", default=False
        ) or api.portal.get_registry_record("plone.email_from_address")
        from_name = api.portal.get_registry_record(
            "euphorie.notification__email_from_name", default=False
        ) or api.portal.get_registry_record("plone.email_from_name")
        return formataddr((from_name, from_address))

    @property
    def recipient(self):
        return formataddr(
            (
                self.webhelpers.get_user_fullname(self.account),
                self.webhelpers.get_user_email(self.account),
            )
        )

    def index(self):
        """The mail text."""
        full_name = self.webhelpers.get_user_fullname(self.account)

        title_missing = api.portal.translate(
            _("label_missing_title", default="Title is missing")
        )

        session_links = ""
        for session in self.sessions:
            session_url = ""
            try:
                session_url = session.absolute_url()
            except ValueError:
                logger.exception(
                    "Could not get session URL for session %r and "
                    "zodb_path %r (session id: %r). There might be a data "
                    "inconsistency. Not including this session in the "
                    "notification mailing.",
                    session.title or title_missing,
                    session.zodb_path,
                    session.id,
                )
                continue

            session_links += (
                f"* [{session.title or title_missing}]({session_url}/@@start)\n"
            )

        # Compile the preferences link
        some_session = self.sessions[0]
        country = api.portal.get().client[some_session.country]
        country_url = country.absolute_url()
        preferences_link = f"{country_url}/preferences"

        return api.portal.translate(
            _(
                "notification_mail_text__base",
                default="""\
Hello ${full_name},

${main_text}

${session_links}

Best regards
Your OiRA Team

**This is an automatically generated mail. If you do not want to receive mails \
from OiRA, you can change this [here](${preferences_link})**""",
                mapping={
                    "full_name": full_name,
                    "main_text": self.main_text,
                    "session_links": session_links,
                    "preferences_link": preferences_link,
                },
            )
        )

    def send_email(self):
        if not self.webhelpers.get_user_email(self.account):
            logger.warning(
                "Account %r has no email address. Not sending notification email.",
                self.account.id,
            )
            return
        super().send_email()


class BaseNotification:
    id = None
    default = False

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @property
    def title_missing(self):
        return api.portal.translate(
            _("label_missing_title", default="Title is missing")
        )

    def notification_enabled(self, account):
        query = (
            Session.query(model.NotificationSubscription)
            .filter(model.NotificationSubscription.account_id == account.id)
            .filter(model.NotificationSubscription.category == self.id)
        )
        result = query.first()
        if result:
            return result.enabled
        return self.default



from euphorie.client import model
from euphorie.client.authentication import EuphorieAccountPlugin
from euphorie.client.browser.login import Login
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.tests.database import DatabaseTests
from Products.PluggableAuthService.interfaces.plugins import IAuthenticationPlugin
from Products.PluggableAuthService.interfaces.plugins import IChallengePlugin
from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin
from Products.PluggableAuthService.interfaces.plugins import IUserEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import IUserFactoryPlugin
from unittest import mock
from unittest import TestCase
from z3c.saconfig import Session
from zope.interface import directlyProvides
from zope.interface.verify import verifyClass


class MockContext:
    def absolute_url(self):
        return "http://www.example.com/base"


class MockRequest:
    PUBLISHED = MockContext()

    def __init__(self, **kw):
        self.__dict__.update(kw)

    def get(self, key, default=None):
        return getattr(self, key, default)

    getHeader = get


class MockResponse:
    def redirect(self, url, lock):
        self.redirect_url = url
        self.redirect_lock = lock


class EuphorieAccountPluginTests(DatabaseTests):
    def test_extraction_interface(self):
        verifyClass(IExtractionPlugin, EuphorieAccountPlugin)

    def test_extraction_no_token_header(self):
        plugin = EuphorieAccountPlugin("plugin")
        request = mock.Mock()
        request.getHeader.return_value = None
        self.assertEqual(plugin.extractCredentials(request), {})

    def test_authenticate_interface(self):
        verifyClass(IAuthenticationPlugin, EuphorieAccountPlugin)

    def test_authenticate_token_no_token(self):
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin._authenticate_token({}), None)

    def test_authenticate_token_call_cms_authenticate(self):
        plugin = EuphorieAccountPlugin("plugin")
        with mock.patch(
            "euphorie.client.authentication.authenticate_cms_token",
            return_value="result",
        ):
            self.assertEqual(plugin._authenticate_token({"api-token": "y"}), "result")

    def test_authenticate_login_wrong_credential_type(self):
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin._authenticate_login({"cookie": "yummie"}), None)

    def test_authenticate_login_unknown_account(self):
        plugin = EuphorieAccountPlugin("plugin")
        credentials = {"login": "login", "password": "secret"}
        self.assertEqual(plugin._authenticate_login(credentials), None)

    def test_authenticate_login_valid_login(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        plugin = EuphorieAccountPlugin("plugin")
        credentials = {"login": "john", "password": "jane"}
        self.assertTrue(plugin._authenticate_login(credentials) is not None)

    def test_authenticate_login_not_case_sensitive(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        plugin = EuphorieAccountPlugin("plugin")
        credentials = {"login": "JoHn", "password": "jane"}
        self.assertTrue(plugin._authenticate_login(credentials) is not None)

    def test_CreateUser_interface(self):
        verifyClass(IUserFactoryPlugin, EuphorieAccountPlugin)

    def test_CreateUser_unknown_account(self):
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.createUser("1", "john"), None)

    def testCreateUser_ValidAccount(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        directlyProvides(request, IClientSkinLayer)
        plugin = EuphorieAccountPlugin("plugin")
        plugin.REQUEST = request
        self.assertTrue(plugin.createUser("1", "john") is account)

    def testEnumerateUsers_Interface(self):
        verifyClass(IUserEnumerationPlugin, EuphorieAccountPlugin)

    def testEnumerateUsers_NoInexactMatch(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.enumerateUsers(login="john", exact_match=False), [])

    def test_EnumerateUsers_search_by_id(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        directlyProvides(request, IClientSkinLayer)
        plugin = EuphorieAccountPlugin("plugin")
        plugin.REQUEST = request
        info = plugin.enumerateUsers(id="1", exact_match=True)
        self.assertEqual(info, [{"id": "1", "login": "john"}])
        self.assertTrue(isinstance(info[0]["id"], str))
        self.assertTrue(isinstance(info[0]["login"], str))

    def test_EnumerateUsers_search_by_login(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        directlyProvides(request, IClientSkinLayer)
        plugin = EuphorieAccountPlugin("plugin")
        plugin.REQUEST = request
        self.assertEqual(
            plugin.enumerateUsers(login="john", exact_match=True),
            [{"id": "1", "login": "john"}],
        )

    def test_EnumerateUsers_search_by_login_and_id(self):
        session = Session()
        account = model.Account(loginname="john", password="jane")
        session.add(account)
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        directlyProvides(request, IClientSkinLayer)
        plugin = EuphorieAccountPlugin("plugin")
        plugin.REQUEST = request
        self.assertEqual(
            plugin.enumerateUsers(id="1", login="john", exact_match=True),
            [{"id": "1", "login": "john"}],
        )

    def test_EnumerateUsers_unknown_account(self):
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.enumerateUsers(id="1", exact_match=False), [])

    def test_Challenge_interface(self):
        verifyClass(IChallengePlugin, EuphorieAccountPlugin)

    def test_Challenge_require_IClientSkinLayer(self):
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        response = MockResponse()
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.challenge(request, response), False)

    def test_Challenge_no_query_string(self):
        request = MockRequest(ACTUAL_URL="http://www.example.com/client")
        directlyProvides(request, IClientSkinLayer)
        response = MockResponse()
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.challenge(request, response), True)
        self.assertEqual(
            response.redirect_url,
            "http://www.example.com/base/@@login?"
            "came_from=http%3A%2F%2Fwww.example.com%2Fclient",
        )
        self.assertEqual(bool(response.redirect_lock), True)

    def test_Challenge_with_query_string(self):
        request = MockRequest(
            ACTUAL_URL="http://www.example.com/client", QUERY_STRING="one=1"
        )
        directlyProvides(request, IClientSkinLayer)
        response = MockResponse()
        plugin = EuphorieAccountPlugin("plugin")
        self.assertEqual(plugin.challenge(request, response), True)
        self.assertEqual(
            response.redirect_url,
            "http://www.example.com/base/@@login?"
            "came_from=http%3A%2F%2Fwww.example.com%2Fclient%3Fone%3D1",
        )


class PasswordPolicyTests(TestCase):
    def setUp(self):
        self.login_view = Login(None, None)

    def test_password_valid(self):
        self.assertTrue(self.login_view.is_valid_password("Abcdef123456"))

    def test_password_length(self):
        self.assertFalse(
            self.login_view.is_valid_password("Ab12"),
            "Minimal length not enforced",
        )

    def test_password_upper_case(self):
        self.assertFalse(
            self.login_view.is_valid_password("abcdef123456"),
            "Upper case letter not enforced",
        )

    def test_password_lower_case(self):
        self.assertFalse(
            self.login_view.is_valid_password("ABCDEF123456"),
            "Lower case letter not enforced",
        )

    def test_password_digit(self):
        self.assertFalse(
            self.login_view.is_valid_password("ABCDEFghijkl"),
            "Digit not enforced",
        )



from euphorie.testing import EUPHORIE_FUNCTIONAL_TESTING
from plone.testing import layered

import doctest
import glob
import os.path
import unittest


def test_suite():
    location = os.path.dirname(__file__) or "."
    doctests = [
        "stories/" + os.path.basename(test)
        for test in glob.glob(os.path.join(location, "stories", "*.txt"))
    ]

    options = (
        doctest.REPORT_ONLY_FIRST_FAILURE
        | doctest.ELLIPSIS
        | doctest.NORMALIZE_WHITESPACE
    )

    suites = [
        layered(
            doctest.DocFileSuite(
                test,
                optionflags=options,
                module_relative=True,
                package="euphorie.client.tests",
                encoding="utf-8",
            ),
            layer=EUPHORIE_FUNCTIONAL_TESTING,
        )
        for test in doctests
    ]
    return unittest.TestSuite(suites)



"""
Sector
======

A sector is a national organisation for a specific type of industry. The
sector content type is a minimal Dexterity item type, using a custom view
template.

Each sector is also a user. This is implemented via the `membrane`
framework.

.. _membrane: http://pypi.python.org/pypi/Products.membrane
"""

from .. import MessageFactory as _
from Acquisition import aq_base
from Acquisition import aq_chain
from Acquisition import aq_inner
from borg.localrole.interfaces import ILocalRoleProvider
from euphorie.content.behaviour.dirtytree import isDirty
from euphorie.content.survey import ISurvey
from euphorie.content.surveygroup import ISurveyGroup
from euphorie.content.user import IUser
from euphorie.content.user import UserProvider
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.indexer import indexer
from plone.namedfile import field as filefield
from plone.supermodel import model
from plone.uuid.interfaces import IAttributeUUID
from plonetheme.nuplone.utils import checkPermission
from Products.CMFCore.utils import getToolByName
from Products.CMFEditions.Permissions import AccessPreviousVersions
from zope import schema
from zope.component import adapter
from zope.interface import implementer

import datetime
import logging


log = logging.getLogger(__name__)


class ISector(model.Schema, IUser, IBasic):
    """Sector object.

    A sector is a national organisation for a specific type of industry.
    """

    directives.order_before(title="*")
    directives.write_permission(title="euphorie.content.ManageCountry")

    directives.order_after(login="title")
    directives.write_permission(login="euphorie.content.ManageCountry")

    directives.order_before(password="description")

    directives.order_after(locked="password")
    directives.write_permission(locked="euphorie.content.ManageCountry")

    contact_name = schema.TextLine(
        title=_("label_contact_name", default="Contact name"), required=True
    )

    directives.order_after(contact_email="contact_name")

    logo = filefield.NamedBlobImage(
        title=_("label_logo", default="Logo"),
        description=_(
            "help_image_upload",
            default=(
                "The logo will appear on the sector overview page of your country. "
                "Make sure your image is of format png, jpg or gif and "
                "does not contain any special characters. "
                "The new logo will only become visible after "
                "you have saved your changes and "
                "published the OiRA tool."
            ),
        ),
        required=False,
    )


@implementer(ISector, INavigationRoot, IAttributeUUID)
class Sector(Container):
    """A sector of industry.

    A sector also acts as a user account in the system, using the
    membrane framework.
    """

    portal_type = "euphorie.clientsector"

    locked = False

    def _canCopy(self, op=0):
        """Tell Zope2 that this object can not be copied."""
        return op


@indexer(ISector)
def SearchableTextIndexer(obj):
    return " ".join(
        [obj.title, obj.description, obj.contact_name or "", obj.contact_email or ""]
    )


@adapter(ISector)
@implementer(ILocalRoleProvider)
class SectorLocalRoleProvider:
    """`borg.localrole` provider for :obj:`ISector` instances.

    This local role provider gives the sector user itself the `Sector`
    local role.
    """

    def __init__(self, context):
        self.context = context

    def getRoles(self, principal_id):
        mt = getToolByName(self.context, "membrane_tool")
        user = mt.getUserObject(user_id=principal_id, brain=True)
        if user is None:
            return ()

        user = user._unrestrictedGetObject()
        if ISector.providedBy(user) and aq_base(user) is aq_base(self.context):
            return ("Sector",)
        return ()

    def getAllRoles(self):
        info = UserProvider(self.context)
        return [(info.getUserId(), ("Sector",))]


def getSurveys(context):
    """Return a list of all surveys for the current sector.

    The return value is a sorted list of dictionaries describing the
    surveygroups for the sector. Each dictionary has the following
    keys:

    * ``id``: surveygroup id
    * ``title`` surveygroup title
    * ``url``: URL for the surveygroup
    * ``published``: boolean indicating if this surveygroup is published
    * ``obsolete``: boolean indicating if this surveygroup is obsolete
    * ``surveys``: list of surveys for the surveygroup. Each entry is a
      dictionary with the following keys:

      * ``id``: survey id
      * ``title``: survey title
      * ``url``: URL for the survey
      * ``published``: boolean indicating if this survey is the currently
        published version of the surveygroup
      * ``current``: boolean indicating if the *context* is inside this survey
      * ``versions``: list of published versions
    """
    current_version = None
    for sector in aq_chain(aq_inner(context)):
        if ISurvey.providedBy(sector):
            current_version = aq_base(sector)
        if ISector.providedBy(sector):
            break
    else:
        return []

    result = []
    groups = [group for group in sector.values() if ISurveyGroup.providedBy(group)]
    repository = getToolByName(context, "portal_repository")
    allow_history = checkPermission(context, AccessPreviousVersions)

    def morph(group, survey):
        published = survey.id == group.published
        info = {
            "id": survey.id,
            "title": survey.title,
            "url": survey.absolute_url(),
            "published": published,
            "publication_date": published and survey.published or None,
            "current": aq_base(survey) is current_version,
            "modified": isDirty(survey),
            "versions": [],
        }
        if not allow_history:
            return info

        history = repository.getHistoryMetadata(survey)
        if history:
            for id in range(history.getLength(countPurged=False) - 1, -1, -1):
                meta = history.retrieve(id, countPurged=False)["metadata"][
                    "sys_metadata"
                ]
                info["versions"].append(
                    {
                        "timestamp": datetime.datetime.fromtimestamp(meta["timestamp"]),
                        "history_id": meta["parent"]["history_id"],
                        "version_id": meta["parent"]["version_id"],
                        "location_id": meta["parent"]["location_id"],
                    }
                )
            info["versions"].sort(key=lambda x: x["timestamp"], reverse=True)
        return info

    for group in groups:
        info = {
            "id": group.id,
            "title": group.title,
            "url": group.absolute_url(),
            "published": bool(group.published),
            "obsolete": bool(group.obsolete),
        }
        info["surveys"] = [
            morph(group, survey)
            for survey in group.values()
            if ISurvey.providedBy(survey)
        ]
        info["surveys"].sort(key=lambda s: s["title"].lower())
        result.append(info)
    result.sort(key=lambda g: g["title"].lower())
    return result



from alembic import command
from alembic import op
from alembic.config import Config
from euphorie.client.model import Session
from logging import getLogger
from pkg_resources import resource_filename
from sqlalchemy import inspect
from zope.deprecation import deprecate


logger = getLogger(__name__)


def has_table(name):
    """Utility function that can be used in alembic upgrades to check if a
    table exists."""
    return name in inspect(op.get_bind()).get_table_names()


def has_column(table_name, column_name):
    """Utility function that can be used in alembic upgrades to check if a
    column exists."""
    bind = op.get_bind()
    columns = inspect(bind).get_columns(table_name)
    return any(column["name"] == column_name for column in columns)


def alembic_upgrade():
    """Upgrade the database to the alembic head."""
    script_location = resource_filename("euphorie.deployment.upgrade", "alembic")
    url = Session().bind.engine.url.__to_string__(hide_password=False)
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", script_location)
    alembic_cfg.set_main_option("sqlalchemy.url", url)
    command.upgrade(alembic_cfg, "head")


@deprecate(
    "Use alembic_upgrade() to reach the head alembic revision",
)
def alembic_upgrade_to(revision):
    script_location = resource_filename("euphorie.deployment.upgrade", "alembic")
    url = Session().bind.engine.url.__to_string__(hide_password=False)
    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", script_location)
    alembic_cfg.set_main_option("sqlalchemy.url", url)
    try:
        command.upgrade(alembic_cfg, revision)
    except Exception:
        logger.exception(
            "Migration failed, you might need to adapt the script to match "
            "your DB state"
        )



from Acquisition import Implicit
from euphorie.client.model import Session
from euphorie.client.model import SessionRedirect
from euphorie.client.model import SurveySession
from euphorie.content.survey import ISurvey
from OFS.Traversable import Traversable
from plone.memoize.instance import memoizedproperty
from sqlalchemy.orm.exc import NoResultFound
from zExceptions import NotFound
from zExceptions import Redirect
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.traversing.namespace import SimpleHandler


class ITraversedSurveySession(Interface):
    """Interface for TraversedSurveySessions."""


@implementer(ITraversedSurveySession)
class TraversedSurveySession(Implicit, Traversable):
    """A traversable session object."""

    def __init__(self, parent, session_id):
        self.__of__(parent)
        self.session_id = int(session_id)
        self.id = f"++session++{self.session_id}"
        self.zodb_path = "/".join(parent.getPhysicalPath()[-3:])

    def getId(self):
        return self.id

    @memoizedproperty
    def session(self):
        try:
            return (
                Session.query(SurveySession)
                .filter(
                    SurveySession.id == self.session_id,
                    SurveySession.zodb_path == self.zodb_path,
                )
                .one()
            )
        except NoResultFound:
            raise NotFound


@adapter(ISurvey, IBrowserRequest)
class SessionTraversal(SimpleHandler):
    factory = TraversedSurveySession

    def get_redirect(self, session_id):
        query = Session.query(SessionRedirect).filter(
            SessionRedirect.old_session_id == session_id
        )
        if query.count() == 0:
            return None
        new_session_id = query.one().new_session_id
        return self.get_redirect(new_session_id) or new_session_id

    def traverse(self, session_id, ignored):
        new_session_id = self.get_redirect(session_id)
        if new_session_id:
            new_traversed_session = self.factory(self.context, new_session_id)
            raise Redirect(
                "/".join(
                    [new_traversed_session.__of__(self.context).absolute_url()]
                    + self.context.REQUEST.path
                )
            )
        return self.factory(self.context, session_id)



try:
    from lxml_html_clean.clean import _find_external_links
    from lxml_html_clean.clean import Cleaner
except ImportError:
    # BBB for lxml 5.1 or earlier, Plone 6.0.10 or earlier
    from lxml.html.clean import _find_external_links
    from lxml.html.clean import Cleaner


marker = []


class LaundryCleaner(Cleaner):
    link_target = marker

    def __call__(self, doc):
        super().__call__(doc)
        if self.link_target is not marker:
            self.force_link_target(doc, self.link_target)

    def force_link_target(self, doc, target):
        for el in _find_external_links(doc):
            if target is None:
                if "target" in el.attrib:
                    del el.attrib["target"]
            elif isinstance(target, (list, tuple)):
                el.set("target", target[0])
            else:
                el.set("target", target)


DocumentCleaner = LaundryCleaner(
    page_structure=False,
    remove_unknown_tags=False,
    allow_tags=[
        "blockquote",
        "a",
        "img",
        "em",
        "p",
        "strong",
        "h3",
        "h4",
        "h5",
        "ul",
        "ol",
        "li",
        "sub",
        "sup",
        "abbr",
        "acronym",
        "dl",
        "dt",
        "dd",
        "cite",
        "dft",
        "br",
        "table",
        "tr",
        "td",
        "th",
        "thead",
        "tbody",
        "tfoot",
    ],
    safe_attrs_only=True,
    add_nofollow=True,
    scripts=True,
    javascript=True,
    comments=False,
    style=True,
    links=False,
    meta=False,
    processing_instructions=False,
    frames=False,
    annoying_tags=False,
)


# Useful for line fields such as titles
LineCleaner = LaundryCleaner(
    page_structure=False,
    safe_attrs_only=True,
    remove_unknown_tags=False,  # Weird API..
    allow_tags=["em", "strong"],
    add_nofollow=True,
    scripts=True,
    javascript=True,
    comments=False,
    style=True,
    processing_instructions=False,
    frames=False,
    annoying_tags=False,
)

CommentCleaner = LaundryCleaner(
    page_structure=False,
    safe_attrs_only=True,
    remove_unknown_tags=False,  # Weird API..
    allow_tags=["blockquote", "a", "em", "p", "strong"],
    add_nofollow=True,
    scripts=False,
    javascript=True,
    comments=False,
    style=True,
    processing_instructions=False,
    frames=False,
    annoying_tags=False,
    link_target=["_blank"],
)


__all__ = ["DocumentCleaner", "LineCleaner", "CommentCleaner"]



from euphorie.content.sector import getSurveys
from euphorie.content.tests.utils import addSurvey
from euphorie.content.tests.utils import createSector
from euphorie.content.tests.utils import EMPTY_SURVEY
from euphorie.content.user import UserAuthentication
from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_NAME
from Products.CMFCore.utils import _checkPermission
from zope.component import getMultiAdapter


class SectorTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.client = self.portal.client
        self.loginAsPortalOwner()

    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createSector(self):
        country = self.portal.sectors.nl
        sector = self._create(country, "euphorie.sector", "sector")
        return sector

    def testNotGloballyAllowed(self):
        types = [fti.id for fti in self.portal.allowedContentTypes()]
        self.assertTrue("euphorie.sector" not in types)

    def testAllowedContentTypes(self):
        sector = self.createSector()
        types = [fti.id for fti in sector.allowedContentTypes()]
        self.assertEqual(set(types), {"euphorie.surveygroup"})

    def testCanNotBeCopied(self):
        sector = self.createSector()
        self.assertFalse(sector.cb_isCopyable())

    def testDeleteWithoutPublishedSurvey(self):
        sector = createSector(self.portal)
        addSurvey(sector, EMPTY_SURVEY)
        surveygroup = sector["test-survey"]
        self.assertEqual(surveygroup.published, None)
        deleteaction = getMultiAdapter((sector, sector.REQUEST), name="delete")
        self.assertEqual(deleteaction.verify(sector.aq_parent, sector), True)

    def testDeleteWithPublishedSurvey(self):
        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        surveygroup = sector["test-survey"]
        publishview = getMultiAdapter((survey, survey.REQUEST), name="publish")
        publishview.publish()
        self.assertEqual(surveygroup.published, "standard-version")
        deleteaction = getMultiAdapter((sector, sector.REQUEST), name="delete")
        self.assertEqual(deleteaction.verify(sector.aq_parent, sector), False)


class SectorAsUserTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createSector(self):
        country = self.portal.sectors.nl
        with api.env.adopt_user("admin"):
            sector = self._create(country, "euphorie.sector", "sector")
        sector.login = "sector"
        sector.indexObject()
        return sector

    def testGetUser(self):
        self.createSector()
        account = self.portal.acl_users.getUser("sector")
        self.assertTrue(account.getUserId())
        self.assertEqual(account.getUserName(), "sector")

    def testGetUserById(self):
        sector = self.createSector()
        uid = sector.UID()
        account = self.portal.acl_users.getUserById(uid)
        self.assertEqual(account.getUserId(), uid)
        self.assertEqual(account.getUserName(), "sector")

    def testGetUserProperties(self):
        sector = self.createSector()
        sector.title = "This is a sector"
        sector.contact_email = "sector@example.com"
        account = self.portal.acl_users.getUser("sector")
        self.assertEqual(account.getProperty("fullname"), "This is a sector")
        self.assertEqual(account.getProperty("email"), "sector@example.com")

    def testSetProperties(self):
        sector = self.createSector()
        sector.title = "This is a sector"
        account = self.portal.acl_users.getUser("sector")
        account.setProperties(fullname="My New Name")
        self.assertEqual(sector.title, "My New Name")

    def testResetPassword(self):
        # Part of https://code.simplon.biz/tracker/tno-euphorie/ticket/111
        sector = self.createSector()
        sector.title = "This is a sector"
        pas = self.portal.acl_users
        auth = UserAuthentication(sector)
        self.assertEqual(auth.authenticateCredentials(dict(password="s3cr3t")), None)
        pas.userSetPassword(auth.getUserId(), "s3cr3t")
        self.assertEqual(
            auth.authenticateCredentials(dict(password="s3cr3t")),
            (auth.getUserId(), auth.getUserName()),
        )


class SectorBrowserTests(EuphorieFunctionalTestCase):
    _default_credentials = {
        "username": SITE_OWNER_NAME,
        "password": SITE_OWNER_PASSWORD,
    }

    def testDuplicateLoginNotAllowed(self):
        # Test for http://code.simplon.biz/tracker/euphorie/ticket/152
        createSector(self.portal, login="sector")
        browser = self.get_browser(logged_in=True)
        browser.open("%s/sectors/nl/@@manage-users" % self.portal.absolute_url())
        browser.getLink("Add new sector").click()
        browser.getControl(name="form.widgets.title").value = "New sector"
        browser.getControl(name="form.widgets.login").value = "sector"
        browser.getControl(name="form.widgets.password").value = "secret"
        browser.getControl(name="form.widgets.password.confirm").value = "secret"
        browser.getControl(name="form.widgets.contact_name").value = "John Doe"
        browser.getControl(name="form.widgets.contact_email").value = "john@example.com"
        browser.getControl(name="form.buttons.save").click()
        self.assertTrue("This login name is already taken" in browser.contents)

    def testPasswordPolicy(self):
        createSector(self.portal, login="sector")
        browser = self.get_browser(logged_in=True)
        browser.open("%s/sectors/nl/@@manage-users" % self.portal.absolute_url())
        browser.getLink("Add new sector").click()
        browser.getControl(name="form.widgets.title").value = "New sector"
        browser.getControl(name="form.widgets.login").value = "sector"
        browser.getControl(name="form.widgets.password").value = "secret"
        browser.getControl(name="form.widgets.password.confirm").value = "secret"
        browser.getControl(name="form.widgets.contact_name").value = "Max Mustermann"
        browser.getControl(name="form.widgets.contact_email").value = "max@example.com"
        browser.getControl(name="form.buttons.save").click()
        self.assertTrue(
            "Your password must contain at least 5 characters, "
            "including at least one capital letter, one number and "
            "one special character (e.g. $, # or @)." in browser.contents
        )

    def testPasswordConfirmation(self):
        createSector(self.portal, login="sector")
        browser = self.get_browser(logged_in=True)
        browser.open("%s/sectors/nl/@@manage-users" % self.portal.absolute_url())
        browser.getLink("Add new sector").click()
        browser.getControl(name="form.widgets.title").value = "New sector"
        browser.getControl(name="form.widgets.login").value = "sector"
        browser.getControl(name="form.widgets.password").value = "S3cr3t!"
        browser.getControl(name="form.widgets.password.confirm").value = (
            "anotherS3cr3t!"
        )
        browser.getControl(name="form.widgets.contact_name").value = "Max Mustermann"
        browser.getControl(name="form.widgets.contact_email").value = "max@example.com"
        browser.getControl(name="form.buttons.save").click()
        self.assertIn(
            "Password doesn't compare with confirmation value", browser.contents
        )


class PermissionTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createSector(self):
        country = self.portal.sectors.nl
        with api.env.adopt_user("admin"):
            sector = self._create(country, "euphorie.sector", "sector")
        sector.login = "sector"
        sector.indexObject()
        return sector

    def testSectorWorkflowUsed(self):
        with api.env.adopt_user(TEST_USER_NAME):
            sector = self.createSector()
        wt = self.portal.portal_workflow
        self.assertEqual(wt.getChainFor(sector), ("sector",))
        self.assertEqual(wt.getInfoFor(sector, "review_state"), "hidden")

    def testAnonymous(self):
        with api.env.adopt_user(TEST_USER_NAME):
            sector = self.createSector()
        self.logout()
        self.assertTrue(not _checkPermission("Euphorie: Add new RIE Content", sector))
        self.assertTrue(not _checkPermission("Access contents information", sector))
        self.assertFalse(_checkPermission("Modify portal content", sector))
        self.assertFalse(_checkPermission("View", sector))

    def testSector(self):
        with api.env.adopt_user(TEST_USER_NAME):
            sector = self.createSector()
        uid = sector.UID()
        user = api.user.get(username=uid)
        permissions = api.user.get_permissions(user=user, obj=sector)
        self.assertTrue(permissions["Euphorie: Add new RIE Content"])
        self.assertTrue(permissions["Access contents information"])
        self.assertTrue(permissions["Modify portal content"])
        self.assertTrue(permissions["View"])

    def testCountryManager(self):
        with api.env.adopt_user(TEST_USER_NAME):
            sector = self.createSector()
        self.portal.acl_users._doAddUser("support", "secret", ["CountryManager"], [])
        self.login("support")
        self.assertTrue(_checkPermission("Euphorie: Add new RIE Content", sector))
        self.assertTrue(_checkPermission("Access contents information", sector))
        self.assertTrue(_checkPermission("Modify portal content", sector))
        self.assertTrue(_checkPermission("View", sector))

    def testManager(self):
        with api.env.adopt_user(TEST_USER_NAME):
            sector = self.createSector()
        self.portal.acl_users._doAddUser("manager", "secret", ["Manager"], [])
        self.login("manager")
        self.assertTrue(_checkPermission("Euphorie: Add new RIE Content", sector))
        self.assertTrue(_checkPermission("Access contents information", sector))
        self.assertTrue(_checkPermission("Modify portal content", sector))
        self.assertTrue(_checkPermission("View", sector))


class GetSurveysTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()

    def getSurveys(self, context):
        return getSurveys(context)

    def testOutsideSector(self):
        self.assertEqual(self.getSurveys(None), [])

    def testSingleUnpublishedSurvey(self):
        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        self.assertEqual(
            self.getSurveys(survey),
            [
                {
                    "url": "http://nohost/plone/sectors/nl/sector/test-survey",
                    "obsolete": False,
                    "published": False,
                    "id": "test-survey",
                    "title": "Test survey",
                    "surveys": [
                        {
                            "id": "standard-version",
                            "title": "Standard version",
                            "current": True,
                            "url": "http://nohost/plone/sectors/nl/sector/"
                            "test-survey/standard-version",
                            "versions": [],
                            "modified": False,
                            "published": False,
                            "publication_date": None,
                        }
                    ],
                }
            ],
        )

    def testTwoUnpublishedSurveysgroups(self):
        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        addSurvey(sector, EMPTY_SURVEY, "Test survey 2")
        result = self.getSurveys(survey)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]["title"], "Test survey")
        self.assertEqual(result[1]["title"], "Test survey 2")

    def testSurveygroupTwoSurveys(self):
        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        surveygroup = sector["test-survey"]
        surveygroup.invokeFactory(
            "euphorie.survey", "next-edition", title="Very latest"
        )
        result = self.getSurveys(survey)
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0]["surveys"]), 2)
        self.assertEqual(result[0]["surveys"][0]["current"], True)
        self.assertEqual(result[0]["surveys"][1]["current"], False)
        self.assertEqual(result[0]["surveys"][1]["title"], "Very latest")

    def testPublishedSurvey(self):
        sector = createSector(self.portal)
        survey = addSurvey(sector, EMPTY_SURVEY)
        surveygroup = sector["test-survey"]
        surveygroup.published = "standard-version"
        result = self.getSurveys(survey)
        self.assertEqual(result[0]["published"], True)
        self.assertEqual(result[0]["surveys"][0]["published"], True)



from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.content.country import ICountry
from euphorie.content.utils import summarizeCountries
from plone.tiles import Tile
from plonetheme.nuplone.skin.interfaces import NuPloneSkin
from plonetheme.nuplone.tiles.navigation import CatalogNavTree
from plonetheme.nuplone.tiles.navigation import INavtreeFactory
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


class _DummyBrain:
    portal_type = None


DummyBrain = _DummyBrain()


@adapter(Interface, NuPloneSkin)
@implementer(INavtreeFactory)
class EuphorieNavtreeFactory:
    """Special navigation tree for the Euphorie surveys.

    This navtree factory modifies the navtree data to remove the survey
    level from the navtree and making the survey contents appear
    directly underneath the surveygroup. This hides versioning
    implementation from the user.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        tree = CatalogNavTree(self.context, self.request)
        walker = tree.iter()
        node = next(walker)

        try:
            while True:
                pt = node.get("brain", DummyBrain).portal_type
                if pt == "euphorie.surveygroup":
                    if not node["ancestor"]:
                        node = walker.send("prune")
                        continue

                    # Cut out the middle man
                    survey = [
                        child
                        for child in node["children"]
                        if child["ancestor"] or child["current"]
                    ]
                    node["children"] = survey[0]["children"]
                node = next(walker)
        except StopIteration:
            pass
        return tree


class UserManagementNavtree(Tile):
    """Special navigation tree for the usermanagement page.

    This tree uses the locale to get the proper names for the countries
    instead of using the titles of the country objects.
    """

    def update(self):
        country_id = self.context.id
        container = aq_parent(aq_inner(self.context))
        self.countries = summarizeCountries(
            container, self.request, country_id, "Euphorie: Manage country"
        )

    def __call__(self):
        if not ICountry.providedBy(self.context):
            return None

        self.update()
        if len(self.countries) < 2:
            return None
        else:
            return self.index()



from euphorie.client import model
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from unittest import mock
from zExceptions import Unauthorized


class TestArchivingViews(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.account = addAccount(password="secret")
        survey_session = model.SurveySession(
            title="Dummy session 1",
            zodb_path="nl/ict/software-development",
            account=self.account,
        )
        model.Session.add(survey_session)
        survey_session = model.SurveySession(
            title="Dummy session 2",
            zodb_path="nl/ict/software-development",
            account=self.account,
        )
        model.Session.add(survey_session)

    def test_archive_session_view(self):
        country = self.portal.client.nl
        traversed_session = country.ict["software-development"].restrictedTraverse(
            "++session++1"
        )
        session = traversed_session.session

        # Check that an unauthorized user cannot archive it
        john = model.Account(loginname="john@example.com")
        model.Session.add(john)
        with api.env.adopt_user(user=john):
            with self._get_view("archive-session", traversed_session) as view:
                with self.assertRaises(Unauthorized):
                    view()

        # By default, webhelpers.use_archive_feature is False, so the archive
        # feature is disabled. Nobody can archive.
        with api.env.adopt_user(user=self.account):
            with self._get_view("archive-session", traversed_session) as view:
                with self.assertRaises(Unauthorized):
                    view()

        # Now, activate the archive feature.
        # Check that an authorized user can archive it
        with mock.patch(
            "euphorie.client.browser.webhelpers.WebHelpers.use_archive_feature",
            return_value=True,
        ):
            with api.env.adopt_user(user=self.account):
                with self._get_view("archive-session", traversed_session) as view:
                    traversed_session
                    self.assertIsNone(session.archived)
                    view()
                    # Now the archival date is set
                    self.assertIsNotNone(session.archived)
                    # Traversing to the view again,
                    # archiving the same session is now prevented
                    with self._get_view(
                        "archive-session", traversed_session
                    ) as view_again:
                        with self.assertRaises(Unauthorized):
                            view_again()
                    # and we are redirected to the session view...
                    self.assertDictEqual(
                        view.request.response.headers,
                        {
                            "location": "http://nohost/plone/client/nl/ict/software-development/++session++1"  # noqa: E501
                        },
                    )
                    # ... or the referer (if specified)
                    view.request.set("HTTP_REFERER", "http://example.com")
                    view.redirect()
                    self.assertDictEqual(
                        view.request.response.headers,
                        {"location": "http://example.com"},
                    )



from euphorie import MessageFactory as _
from euphorie.client.model import Account
from euphorie.client.utils import CreateEmailTo
from logging import getLogger
from plone import api
from plone.autoform.form import AutoExtensibleForm
from plone.memoize.view import memoize_contextless
from plone.schema import Email
from plone.supermodel import model
from Products.CMFPlone.PasswordResetTool import InvalidRequestError
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.MailHost.MailHost import MailHostError
from requests.utils import is_ipv4_address
from urllib.parse import urlencode
from z3c.form import button
from z3c.form.form import EditForm
from z3c.saconfig import Session
from zope import schema
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Invalid
from zope.interface import invariant
from zope.publisher.interfaces import IPublishTraverse

import smtplib


logger = getLogger(__name__)


class ResetPasswordFormSchema(model.Schema):
    new_password = schema.Password(
        title=_("label_new_password", default="Desired password"),
    )
    new_password_confirmation = schema.Password(
        title=_("label_confirm_password", default="Confirm your password"),
    )

    @invariant
    def validate_same_value(data):
        if data.new_password != data.new_password_confirmation:
            raise Invalid(_("Password doesn't compare with confirmation value"))


class ResetPasswordRequestSchema(model.Schema):
    email = Email(
        title=_("label_email", default="Email address"),
    )


class BaseForm(AutoExtensibleForm, EditForm):
    """Base class for password the reset forms."""

    title = ""
    description = ""

    @property
    def template(self):
        return self.index

    def redirect(self, target, msg="", msg_type="notice"):
        """Redirect the user to a meaningfull place and add a status
        message."""
        if msg:
            api.portal.show_message(msg, self.request, msg_type)
        self.request.response.redirect(target)

    def updateActions(self):
        """Fix the button classes."""
        super().updateActions()
        for action in self.actions.values():
            action.klass = "pat-button"


class ResetPasswordRequest(BaseForm):
    """Request a link to reset the password."""

    ignoreContext = True
    schema = ResetPasswordRequestSchema

    label = _(
        "title_reset_password_request",
        default="Password recovery",
    )
    description = _(
        "description_reset_password_request",
        ("We will send you an email " "with the instructions to reset your password."),
    )
    button_label = _(
        "label_send_password_reminder",
        default="Send password reminder",
    )

    email_template = ViewPageTemplateFile("templates/password_recovery_email.pt")

    @property
    def email_from_name(self):
        return api.portal.get_registry_record("plone.email_from_name")

    @property
    def email_from_address(self):
        return api.portal.get_registry_record("plone.email_from_address")

    def expiration_timeout(self):
        ppr = api.portal.get_tool("portal_password_reset")
        timeout = ppr.getExpirationTimeout() or 0
        return int(timeout * 24)  # timeout is in days, but templates want in hours.

    def log_error(self, msg):
        """Log an error message, set the view error attribute and return
        False."""
        logger.error(msg)
        self.error = _(
            "An error occured while sending the password reset instructions",
        )
        return False

    def get_remote_host(self):
        forwarded_for = self.request.get("HTTP_X_FORWARDED_FOR")
        if is_ipv4_address(forwarded_for):
            return forwarded_for
        return self.request.get("REMOTE_ADDR")

    def send_mail(self, email):
        account = Session.query(Account).filter(Account.loginname == email).first()
        if not account:
            # We returned True even if the account
            # does not exist to not leak any information
            return True

        ppr = api.portal.get_tool("portal_password_reset")
        # Clean out previous requests by this user
        for token, value in ppr._requests.items():
            if value[0] == account.id:
                del ppr._requests[token]
                ppr._p_changed = 1

        reset_info = ppr.requestReset(account.id)
        reset_info["host"] = self.get_remote_host()
        mailhost = api.portal.get_tool("MailHost")
        body = self.email_template(**reset_info)
        subject = translate(
            _("OiRA password reset instructions"),
            context=self.request,
        )
        mail = CreateEmailTo(
            self.email_from_name,
            self.email_from_address,
            account.email,
            subject,
            body,
        )

        try:
            mailhost.send(
                mail,
                account.email,
                self.email_from_address,
                immediate=True,
            )
            logger.info(
                "Sent password reset instructions to %s",
                account.email,
            )
        except MailHostError as e:
            msg = (
                "MailHost error sending password reset instructions to {}: {}"
            ).format(account.email, e)
            return self.log_error(msg)
        except smtplib.SMTPException as e:
            msg = (
                "smtplib error sending password reset instructions to {}: {}"
            ).format(account.email, e)
            return self.log_error(msg)
        except OSError as e:
            msg = ("Socket error sending password reset instructions to {}: {}").format(
                account.email, e[1]
            )
            return self.log_error(msg)
        return True

    def do_next(self):
        data, error = self.extractData()
        if error:
            return
        email = data.get("email")
        if not self.send_mail(email):
            return
        msg = _(
            "message_password_recovery_email_sent",
            default="An email will be sent to ${email} "
            "if we can find an account for this email address. Please use the "
            "link inside the e-mail to reset your password.",
            mapping={"email": email},
        )
        webhelpers = api.content.get_view(
            name="webhelpers", context=self.context, request=self.request
        )
        redir_url = webhelpers.get_came_from(default=self.context.absolute_url())
        if not redir_url.endswith("login"):
            redir_url = "{}/@@login?{}#login".format(
                redir_url, urlencode({"came_from": redir_url})
            )
        self.redirect(redir_url, msg)

    @button.buttonAndHandler(_("Save"))
    def next_handler(self, action):
        """Check if the security token is correct and if it is change the
        account password with the provided value."""
        self.do_next()

    @button.buttonAndHandler(_("Cancel"))
    def handleCancel(self, action):
        self.redirect(self.context.absolute_url())


@implementer(IPublishTraverse)
class ResetPasswordForm(BaseForm):
    ignoreContext = True
    schema = ResetPasswordFormSchema

    label = _(
        "title_reset_password_form",
        default="Reset password",
    )
    description = _(
        "description_reset_password_form",
        default="",
    )
    button_label = _("Save changes")

    def update(self):
        super().update()
        key = self.key
        ppr = api.portal.get_tool("portal_password_reset")
        try:
            ppr.verifyKey(key)
        except InvalidRequestError:
            self.error = _("Invalid security token, try to request a new one")

    def publishTraverse(self, request, name):
        return self

    @property
    @memoize_contextless
    def key(self):
        """Extract the key from the URL."""
        return self.request.getURL().rpartition("/")[-1]

    def do_save(self):
        """Execute the save action."""
        (data, errors) = self.extractData()
        if errors:
            for err in errors:
                if isinstance(err.error, Exception):
                    self.error = _(
                        "error_password_mismatch", default="Passwords do not match"
                    )
            return

        key = self.key
        ppr = api.portal.get_tool("portal_password_reset")

        try:
            ppr.verifyKey(key)
        except InvalidRequestError:
            self.error = _("Invalid security token, try to request a new one")
            return

        login_view = api.content.get_view(
            name="login",
            context=self.context,
            request=self.request,
        )
        error = login_view.check_password_policy(data["new_password"])
        if error:
            self.error = error
            return

        account_id, expiry = ppr._requests.get(key)
        if ppr.expired(expiry):
            del ppr._requests[key]
            ppr._p_changed = 1
            self.error = _("This URL has expired, try to request a new one")
            return

        account = Session().query(Account).filter(Account.id == account_id).one()
        account.password = data["new_password"]

        # clean out the request
        del ppr._requests[key]
        ppr._p_changed = 1

        current_url = self.context.absolute_url()
        return self.redirect(
            f"{current_url}/@@login?{urlencode(dict(came_from=current_url))}",
            msg=_("Your password was successfully changed."),
        )

    @button.buttonAndHandler(_("Save"))
    def save_handler(self, action):
        """Check if the security token is correct and if it is change the
        account password with the provided value."""
        self.do_save()

    @button.buttonAndHandler(_("Cancel"))
    def handleCancel(self, action):
        self.redirect(self.context.absolute_url())



from euphorie.client.model import SurveyTreeItem
from sqlalchemy import and_
from sqlalchemy import Integer
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import cast
from z3c.saconfig import Session


def handle_custom_risks_order(context, event):
    session = Session()

    custom_risks = session.query(SurveyTreeItem).filter(
        and_(
            SurveyTreeItem.session_id == context.session_id,
            SurveyTreeItem.path.like(context.path + "%"),
            SurveyTreeItem.type == "risk",
        )
    )

    # First, set all paths and zodb_paths to bogus values (extra zeros)
    # to avoid constraint errors
    for risk in custom_risks:
        risk.path = risk.path + "000"
        risk.zodb_path = risk.zodb_path + "000"

    ordered_custom_risks = (
        session.query(SurveyTreeItem)
        .filter(
            and_(
                SurveyTreeItem.session_id == context.session_id,
                SurveyTreeItem.path.like(context.path + "%"),
                SurveyTreeItem.type == "risk",
            )
        )
        .order_by(cast(func.split_part(SurveyTreeItem.zodb_path, "/", 2), Integer))
    )

    # Iterate over the risks in their natural order. Close any gaps in numbering
    for count, risk in enumerate(ordered_custom_risks):
        risk.zodb_path = f"custom-risks/{count + 1}"

    # Now, set the path according to the zodb_path (= natural order)
    for risk in custom_risks:
        risk.path = "%s%03d" % (context.path, int(risk.zodb_path.split("/")[-1]))



from AccessControl.Permissions import manage_users as ManageUsers
from euphorie.client import authentication
from Products.PluggableAuthService.PluggableAuthService import registerMultiPlugin
from zope.i18nmessageid import MessageFactory as mf

import mimetypes


mimetypes.add_type("image/svg+xml", ".svg")


MessageFactory = mf("euphorie")
del mf

#: Version number for terms and conditions. Updating this version will
#: require all users to (re)confirm their acceptance of the terms and
#: conditions
CONDITIONS_VERSION = 1


def initialize(context):
    registerMultiPlugin(authentication.EuphorieAccountPlugin.meta_type)
    context.registerClass(
        authentication.EuphorieAccountPlugin,
        permission=ManageUsers,
        constructors=(
            authentication.manage_addEuphorieAccountPlugin,
            authentication.addEuphorieAccountPlugin,
        ),
        visibility=None,
    )

    # Instruct the email module to use quoted printable for UT8
    import email.charset

    email.charset.add_charset("utf-8", email.charset.QP, email.charset.QP, "utf-8")

    # Monkeypatch the publisher to disable its WebDAV logic so we can use
    # all request methods for our REST API.
    import ZPublisher.HTTPRequest

    ZPublisher.HTTPRequest.HTTPRequest.maybe_webdav_client = 0



"""Add consultancy table

Revision ID: 20230421141443
Revises: 20230406070728
Create Date: 2023-04-21 14:17:25.526866

"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20230421141443"
down_revision = "20230406070728"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("consultancy"):
        op.create_table(
            "consultancy",
            sa.Column("session_id", sa.Integer(), nullable=False),
            sa.Column("account_id", sa.Integer(), nullable=False),
            sa.ForeignKeyConstraint(
                ["account_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.ForeignKeyConstraint(
                ["session_id"], ["session.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
        )


def downgrade():
    if has_table("consultancy"):
        op.drop_table("consultancy")



from euphorie.content.behaviour.richdescription import Description

import unittest


class Mock:
    pass


class RichDescriptionTests(unittest.TestCase):
    def testMissing(self):
        self.assertEqual(Description(object())(), "")

    def testEmpty(self):
        obj = Mock()
        obj.description = ""
        self.assertEqual(Description(obj)(), "")

    def testSimpleText(self):
        obj = Mock()
        obj.description = "Test"
        self.assertEqual(Description(obj)(), "Test")

    def testMarkup(self):
        obj = Mock()
        obj.description = "Test <em>me</me> <strong>now</strong>"
        self.assertEqual(Description(obj)(), "Test me now")



from ftw.upgrade import UpgradeStep


class RemoveCloneFeatureFlag(UpgradeStep):
    """Remove clone feature flag, which is always on now."""

    def __call__(self):
        self.install_upgrade_profile()



import unittest
import z3c.saconfig.tests


z3c.saconfig.tests  # Keep PyFlakes happy


class DatabaseTests(unittest.TestCase):
    create_tables = True

    def setUp(self):
        super().setUp()
        from euphorie.client import model
        from Products.Five import fiveconfigure
        from z3c.saconfig import Session
        from Zope2.App.zcml import load_config

        import euphorie.client.tests

        fiveconfigure.debug_mode = True
        load_config("configure.zcml", euphorie.client.tests)
        fiveconfigure.debug_mode = False
        if self.create_tables:
            model.metadata.create_all(Session.bind, checkfirst=True)

    def tearDown(self):
        super().tearDown()
        from euphorie.client import model
        from z3c.saconfig import Session

        import transaction

        Session.remove()
        model.metadata.drop_all(Session.bind)
        transaction.abort()



"""Add the training table.

Revision ID: 20220318135716
Revises: 20211019165915
Create Date: 2022-03-18 14:06:17.384922
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_table

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20220318135716"
down_revision = "20211019165915"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("training"):
        op.create_table(
            "training",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("time", sa.DateTime(), nullable=True),
            sa.Column("account_id", sa.Integer(), nullable=False),
            sa.Column("session_id", sa.Integer(), nullable=False),
            sa.Column("answers", sa.Unicode(), nullable=True),
            sa.Column("status", sa.Unicode(), nullable=True),
            sa.ForeignKeyConstraint(
                ["account_id"], ["account.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.ForeignKeyConstraint(
                ["session_id"], ["session.id"], onupdate="CASCADE", ondelete="CASCADE"
            ),
            sa.PrimaryKeyConstraint("id"),
        )


def downgrade():
    op.drop_table("training")



from zope.deferredimport import deprecated


deprecated(
    (
        "It will be removed in future versions. "
        "Please use the regular image widget instead."
    ),
    LogoWidget="euphorie.content.widgets.logo_bbb:LogoWidget",
    LogoFieldWidget="euphorie.content.widgets.logo_bbb:LogoFieldWidget",
)



from euphorie.client import model
from euphorie.content.behaviour.uniqueid import get_next_id
from euphorie.content.solution import ISolution
from plone.app.upgrade.utils import loadMigrationProfile
from Products.CMFCore.utils import getToolByName
from sqlalchemy.engine.reflection import Inspector
from z3c.saconfig import Session
from zope.sqlalchemy import datamanager

import logging


log = logging.getLogger(__name__)


def add_column_for_training_notes(context):
    session = Session()
    inspector = Inspector.from_engine(session.bind)
    columns = [c["name"] for c in inspector.get_columns(model.Risk.__table__.name)]
    if "training_notes" not in columns:
        log.info("Adding training_notes column for risks")
        session.execute(
            "ALTER TABLE %s ADD training_notes TEXT" % model.Risk.__table__.name
        )
        datamanager.mark_changed(session)


def update_nav_types_registry(context):
    """There is a new registry record called plone.displayed_types that
    determines what types get shown in the navigation."""
    loadMigrationProfile(context, "profile-euphorie.deployment.upgrade:to_0020")


def migrate_existing_measures(context):
    ct = getToolByName(context, "portal_catalog")
    risks = ct(portal_type="euphorie.risk")
    count = 0
    for brain in risks:
        risk = brain.getObject()
        existing_measures = getattr(risk, "existing_measures", None)
        if existing_measures:
            solutions = [
                solution.description.strip()
                for solution in risk.values()
                if ISolution.providedBy(solution)
            ]
            for measure in existing_measures.splitlines():
                if measure.strip() not in solutions:
                    id = get_next_id(risk)
                    risk.invokeFactory("euphorie.solution", id)
                    solution = getattr(risk, id)
                    solution.description = measure
                    solution.action_plan = measure
                    count += 1
    log.info("Created %d Solutions", count)


def extend_zodb_path_field(context):
    session = Session()
    session.execute(
        "ALTER TABLE %s ALTER COLUMN zodb_path TYPE varchar(512);"
        % model.SurveySession.__table__.name
    )
    datamanager.mark_changed(session)



from euphorie.content import MessageFactory as _
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from Products.membrane.config import TOOLNAME
from Products.membrane.interfaces import user as user_ifaces
from Products.statusmessages.interfaces import IStatusMessage


class PasswordPolicyTests(EuphorieIntegrationTestCase):
    def testPasswordPolicy(self):
        err_msg = _(
            "password_policy_conditions",
            default="Your password must contain at least 5 characters, "
            "including at least one capital letter, one number and "
            "one special character (e.g. $, # or @).",
        )
        regtool = api.portal.get_tool("portal_registration")
        self.assertEqual(regtool.pasValidation("password", "secret"), err_msg)
        self.assertEqual(regtool.pasValidation("password", "Secret"), err_msg)
        self.assertEqual(regtool.pasValidation("password", "Secret1"), err_msg)
        self.assertIsNone(regtool.pasValidation("password", "Secret1!"))

    def testLockoutPolicy(self):
        sector = createSector(self.portal)
        mbtool = api.portal.get_tool(TOOLNAME)
        member = mbtool.getUserObject(login=sector.login)
        auth = user_ifaces.IMembraneUserAuth(member, None)
        status = IStatusMessage(auth.context.REQUEST)

        auth.applyLockoutPolicy(0)
        self.assertEqual(len(status.show()), 0)
        self.assertFalse(sector.locked)

        auth.applyLockoutPolicy(2)
        self.assertEqual(len(status.show()), 1)
        self.assertEqual(auth.context._v_login_attempts, 1)
        self.assertFalse(sector.locked)

        auth.applyLockoutPolicy(2)
        self.assertEqual(len(status.show()), 1)
        self.assertEqual(auth.context._v_login_attempts, 2)
        self.assertTrue(sector.locked)



from euphorie.testing import EuphorieFunctionalTestCase
from plone.app.redirector.interfaces import IFourOhFourView
from plone.app.redirector.interfaces import IRedirectionStorage
from plone.contentrules.engine.interfaces import IRuleAssignable
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface.verify import verifyObject


class SimplifiedPloneTests(EuphorieFunctionalTestCase):
    def testRedirectStorageRemoved(self):
        self.assertTrue(queryUtility(IRedirectionStorage) is None)

    def testRedirectViewStillValid(self):
        view = queryMultiAdapter(
            (self.portal, self.portal.REQUEST), name="plone_redirector_view"
        )
        self.assertTrue(verifyObject(IFourOhFourView, view))

    # XXX Disabled for now: in Plone 4 the test setup loads all Plone zcml before
    # the Euphorie test layer is setup, so the <exclude> directories in the
    # euphorie.deployment zcml have no effect.

    def XtestNoContentRulesRegistered(self):
        # Testing for event subscribers would be the cleanest approach, but is
        # hard to do. Instead we test for extra marker interfaces set by the
        # plone.app.contentrules zcml.
        self.assertTrue(not IRuleAssignable.providedBy(self.portal))

    def testNoContentRulesAction(self):
        action = self.portal.portal_actions.object.contentrules
        self.assertEqual(action.visible, False)



from plone import api
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from Products.Five import BrowserView

import markdown


class SolutionView(BrowserView):
    def render_md(self, text):
        md_text = markdown.markdown(text)
        transforms = api.portal.get_tool("portal_transforms")
        data = transforms.convertTo("text/x-html-safe", md_text, mimetype="text/html")
        return data.getData()


class AddForm(DefaultAddForm):
    portal_type = "euphorie.solution"

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["action_plan"].mode = "hidden"
        self.widgets["prevention_plan"].mode = "hidden"
        self.widgets["action"].rows = 15


class AddView(DefaultAddView):
    form = AddForm


class EditForm(DefaultEditForm):
    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["action_plan"].mode = "hidden"
        self.widgets["prevention_plan"].mode = "hidden"
        self.widgets["action"].rows = 15



from setuptools import find_packages
from setuptools import setup


version = "19.1.4"

setup(
    name="Euphorie",
    version=version,
    description="Euphorie Risk Assessment tool",
    long_description=open("README.rst").read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.1",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="euphorie OiRA Interactive Risk Assessment",
    author="Wichert Akkerman and Syslab.com",
    author_email="info@syslab.com",
    url="https://euphorie.readthedocs.org/en/latest/",
    license="GPL",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.11",
    install_requires=[
        "AccessControl",
        "alembic",
        "anytree",
        "Chameleon",
        "decorator",
        "ftw.upgrade",
        "lxml",
        "markdownify",
        "nltk",
        "NuPlone >= 4.0.1",
        "openpyxl",
        "path.py",
        "Plone >=6.0",
        "plone.api",
        "plone.app.imagecropping",
        "plone.app.redirector",
        "plone.app.vocabularies",
        "plone.app.z3cform",
        "plone.behavior",
        "plone.dexterity",
        "plone.formwidget.namedfile",
        "plone.memoize",
        "plone.namedfile",
        "plone.patternslib",
        "plonestatic.euphorie",
        "plone.uuid",
        "Products.membrane >=4.999999",
        "py-bcrypt",
        "python-docx",
        "repoze.formapi",
        "scikit-learn",
        "setuptools",
        "sh",
        "SQLAlchemy >=1.2.999999",
        "stoneagehtml",
        "user_agents",
        "weasyprint",
        "z3c.form",
        "z3c.saconfig",
        "zope.configuration",
        "zope.i18nmessageid",
        "zope.interface",
        "zope.schema",
    ],
    tests_require=[
        "plone.app.testing",
    ],
    extras_require={
        "sphinx": [
            "Sphinx >=1.0",
            "repoze.sphinx.autointerface",
        ],
        "tests": [
            "plone.app.testing",
        ],
    },
    entry_points="""
      [z3c.autoinclude]
      target = plone

      [zopectl.command]
      initdb = euphorie.deployment.commands.initdb:main
      xmlimport = euphorie.deployment.commands.xmlimport:main

      [console_scripts]
      upgradedb = euphorie.deployment.commands.upgradedb:main
      """,
)



from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.model import NotificationSubscription
from euphorie.client.model import SurveySession
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.client.tests.utils import MockMailFixture
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from unittest import mock
from z3c.saconfig import Session
from zope.interface import alsoProvides

import datetime


class NotificationsSettingsTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.account = addAccount(password="secret")
        alsoProvides(self.request, IClientSkinLayer)
        api.portal.set_registry_record("euphorie.notifications__enabled", True)

    def test_all_notifications(self):
        view = api.content.get_view(
            context=self.portal.client.nl, request=self.request, name="preferences"
        )
        with mock.patch(
            "euphorie.client.notifications.notification__ra_not_modified.Notification"
            ".available",
            return_value=True,
        ):
            notification_ids = [category.id for category in view.all_notifications]
        self.assertEqual(notification_ids, ["notification__ra_not_modified"])

    def test_subscribe_notification(self):
        request = self.request.clone()
        request.form["notifications"] = {
            "notification__ra_not_modified": True,
            "form.buttons.save": "",
        }
        view = api.content.get_view(
            context=self.portal.client.nl, request=request, name="preferences"
        )
        with api.env.adopt_user(user=self.account):
            view.update()
            with mock.patch(
                "euphorie.client.notifications.notification__ra_not_modified"
                ".Notification.available",
                return_value=True,
            ):
                view.handleSave(view, None)
        subscriptions = (
            Session.query(NotificationSubscription)
            .filter(NotificationSubscription.account_id == self.account.getId())
            .all()
        )
        self.assertEqual(subscriptions[0].category, "notification__ra_not_modified")

    def test_disallow_user_settings(self):
        api.portal.set_registry_record(
            "euphorie.notifications__allow_user_settings", False
        )
        request = self.request.clone()
        request.form["notifications"] = {
            "notification__ra_not_modified": True,
            "form.buttons.save": "",
        }
        view = api.content.get_view(
            context=self.portal.client.nl, request=request, name="preferences"
        )
        with api.env.adopt_user(user=self.account):
            view.update()

            # This attribute, based on the
            # `euphorie.notifications__allow_user_settings` registry entry,
            # should be set to False
            self.assertEqual(view.allow_notification_settings, False)

            # Try to save the notification
            with mock.patch(
                "euphorie.client.notifications.notification__ra_not_modified"
                ".Notification.available",
                return_value=True,
            ):
                view.handleSave(view, None)

        subscriptions = (
            Session.query(NotificationSubscription)
            .filter(NotificationSubscription.account_id == self.account.getId())
            .all()
        )

        # Notification should not be saved
        self.assertEqual(subscriptions, [])


class NotificationsSendingTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.account = addAccount(password="secret")
        alsoProvides(self.request, IClientSkinLayer)

        survey_session = SurveySession(
            id=1,
            title="Euphorie",
            zodb_path="nl/ict/software-development",
            account=self.account,
        )
        survey_session.modified = datetime.datetime.now() - datetime.timedelta(days=366)
        Session.add(survey_session)
        survey_session2 = SurveySession(
            id=2,
            title="Depublished",
            zodb_path="nl/ict/depublished",
            account=self.account,
        )
        survey_session2.modified = datetime.datetime.now() - datetime.timedelta(
            days=366
        )
        Session.add(survey_session2)
        api.portal.set_registry_record("euphorie.notifications__enabled", True)

    def test_send_notification(self):
        Session.add(
            NotificationSubscription(
                account_id=self.account.getId(),
                category="notification__ra_not_modified",
                enabled=True,
            )
        )
        view = api.content.get_view(
            context=self.portal.client.nl,
            request=self.request,
            name="send-notifications-daily",
        )
        mail_fixture = MockMailFixture()
        with mock.patch(
            "euphorie.client.notifications.notification__ra_not_modified"
            ".Notification.available",
            return_value=True,
        ):
            view()
        mail = mail_fixture.storage[0][0][0]
        self.assertIn(self.account.email, mail.get("To"))
        self.assertEqual(
            mail_fixture.storage[0][1]["subject"],
            "Reminder: Update of risk assessment (1 open)",
        )



from euphorie.content import utils
from euphorie.content.solution import Solution
from euphorie.testing import EuphorieIntegrationTestCase
from zope.component.hooks import getSite
from zope.i18n import translate


def _create(container, *args, **kwargs):
    newid = container.invokeFactory(*args, **kwargs)
    return getattr(container, newid)


def createSolution(algorithm="kinney"):
    portal = getSite()
    country = portal.sectors.nl
    sector = _create(country, "euphorie.sector", "sector")
    surveygroup = _create(
        sector, "euphorie.surveygroup", "group", evaluation_algorithm=algorithm
    )
    survey = _create(surveygroup, "euphorie.survey", "survey")
    module = _create(survey, "euphorie.module", "module")
    risk = _create(module, "euphorie.risk", "risk")
    solution = _create(risk, "euphorie.solution", "solution")
    return solution


class RiskTests(EuphorieIntegrationTestCase):
    def testNotGloballyAllowed(self):
        types = [fti.id for fti in self.portal.allowedContentTypes()]
        self.assertTrue("euphorie.solution" not in types)

    def testCanBeCopied(self):
        self.loginAsPortalOwner()
        solution = createSolution()
        self.assertTrue(solution.cb_isCopyable())

    def testCMFAccessors(self):
        """CMF-Style accessors must return utf8-encoded strings."""
        self.loginAsPortalOwner()
        solution = createSolution()
        title = solution.Title()
        self.assertIsInstance(title, str)
        survey = utils.getSurvey(solution)
        self.assertEqual(
            title, translate(Solution.title, target_language=survey.language)
        )



from euphorie.content import MessageFactory as _
from euphorie.content.surveygroup import ISurveyGroup
from io import StringIO
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView

import csv
import datetime


class SimilarTitlesDetails(BrowserView):
    """A view that shows the risks that have a title similar to the title
    of the current risk.
    """

    @property
    @memoize
    def tool_cache(self):
        return {}

    def get_tool_for_brain(self, risk):
        cache_key = "/".join(risk.getPhysicalPath()[:6])
        cache = self.tool_cache
        value = cache.get(cache_key)
        if value is not None:
            return value
        for obj in risk.aq_chain:
            if ISurveyGroup.providedBy(obj):
                value = obj
                cache[cache_key] = value
                return value

    def label(self):
        return _("Risks similar to ${title}", mapping={"title": self.context.Title()})

    def get_solutions(self, risk):
        return [
            obj for obj in risk.objectValues() if obj.portal_type == "euphorie.solution"
        ]

    def solutions_by_risk(self):
        risks = [self.context] + [
            api.content.get(path) for path in self.request.form.get("paths", []) or []
        ]
        return {risk: self.get_solutions(risk) for risk in risks}


class SimilarTitlesDetailsCSV(SimilarTitlesDetails):
    def write_data(self, data, writer):
        """Write data to CSV writer.
        First cell is the risk title (heading).
        Starting from the second row, every column is headed by a tool title
        and contains one solution per row.
        """
        writer.writerow([self.context.Title()])
        solutions = [
            [self.get_tool_for_brain(risk).Title()]
            + [solution.Description() for solution in solutions]
            for risk, solutions in data.items()
        ]
        max_len = max(map(len, solutions))
        for idx in range(max_len):
            writer.writerow((sub[idx] if len(sub) > idx else "") for sub in solutions)

    def __call__(self):
        buffer = StringIO()
        writer = csv.writer(buffer, delimiter=";")

        self.write_data(self.solutions_by_risk(), writer)

        csv_data = buffer.getvalue()
        buffer.close()
        response = self.request.RESPONSE
        today_iso = datetime.date.today().isoformat()
        response.setHeader(
            "Content-Disposition",
            (
                f"attachment; filename=similar_titles_details_{self.context.Title()}_"
                f"{today_iso}.csv"
            ),
        )
        response.setHeader("Content-Type", "text/csv;charset=utf-8")
        return csv_data



from ftw.upgrade import UpgradeStep
from plone import api


class AddTheLockingFeature(UpgradeStep):
    """Add the locking feature."""

    def __call__(self):
        self.install_upgrade_profile()
        old_value = api.portal.get_registry_record(
            "euphorie.use_publication_feature", default=False
        )
        api.portal.set_registry_record("euphorie.use_locking_feature", old_value)



from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.content import MessageFactory as _
from euphorie.content.interfaces import SurveyUnpublishEvent
from euphorie.content.survey import ISurvey
from euphorie.content.widgets.survey_source_selection import SurveySourceSelectionField
from OFS.event import ObjectClonedEvent
from plone import api
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.utils import createContentInContainer
from plone.uuid.handlers import addAttributeUUID
from Products.CMFCore.interfaces import ISiteRoot
from Products.Five import BrowserView
from urllib.parse import urlencode
from z3c.form.field import Fields
from ZODB.POSException import ConflictError
from zope.component import adapter
from zope.component import getUtility
from zope.event import notify
from zope.interface import implementer
from zope.interface import Interface
from zope.lifecycleevent import ObjectCopiedEvent

import datetime
import logging


log = logging.getLogger(__name__)


class SurveyGroupView(BrowserView):
    def surveys(self):
        templates = [
            dict(title=survey.title, url=survey.absolute_url())
            for survey in self.context.values()
            if ISurvey.providedBy(survey)
        ]
        return templates


class ISurveySourceSelectionSchema(Interface):
    source_selection = SurveySourceSelectionField(
        title=_("label_source_selection", default="Choose a source"),
        required=False,
    )


@implementer(ISurveySourceSelectionSchema)
@adapter(Interface)
class SurveySourceSelectionAdaptedContext:
    def __init__(self, context):
        pass


class AddForm(DefaultAddForm):
    """This add form allows you to create a new OiRA Tool from a template"""

    description = _(
        "intro_add_surveygroup",
        default="This form will allow you to create a new OiRA Tool.",
    )

    portal_type = "euphorie.surveygroup"

    def updateFields(self):
        super().updateFields()
        self.fields = self.fields.omit("description", "obsolete")
        # Add a field with a radio widget to select a choice
        self.fields += Fields(ISurveySourceSelectionSchema)
        try:
            self.fields.move_to_end("evaluation_algorithm")
        except AttributeError:
            self.fields = self.fields.omit("evaluation_algorithm") + self.fields.select(
                "evaluation_algorithm"
            )

    def update(self):
        super().update()
        sector = aq_inner(self.context)
        country = aq_parent(sector)
        self.my_country = country.id
        self.my_sector = sector.id

    def buildSurveyTree(self):
        site = getUtility(ISiteRoot)
        catalog = api.portal.get_tool("portal_catalog")
        startPath = "%s/sectors" % "/".join(site.getPhysicalPath())
        startPathDepth = startPath.count("/") + 1
        query = {
            "path": {"query": startPath, "depth": 4, "navtree": True},
            "portal_type": [
                "euphorie.country",
                "euphorie.sector",
                "euphorie.surveygroup",
                "euphorie.survey",
            ],
            "sort_on": "path",
            "sort_order": "asc",
        }
        tree = {}
        for brain in catalog.searchResults(query):
            path = brain.getPath().split("/")[startPathDepth:]
            if brain.portal_type == "euphorie.country":
                tree[brain.id] = dict(
                    id=brain.id, title=brain.Title, path=brain.getPath(), sectors={}
                )
            elif brain.portal_type == "euphorie.sector":
                country = tree[path[0]]
                country["sectors"][brain.id] = dict(
                    id=brain.id, title=brain.Title, path=brain.getPath(), groups={}
                )
            elif brain.portal_type == "euphorie.surveygroup":
                country = tree[path[0]]
                sector = country["sectors"][path[1]]
                sector["groups"][brain.id] = dict(
                    id=brain.id, title=brain.Title, path=brain.getPath(), surveys={}
                )
            elif brain.portal_type == "euphorie.survey":
                country = tree[path[0]]
                sector = country["sectors"][path[1]]
                group = sector["groups"][path[2]]
                group["surveys"][brain.id] = dict(
                    id=brain.id,
                    title=brain.Title,
                    path=brain.getPath(),
                    url=brain.getURL(),
                )
        my_sector_path = tree[self.my_country]["sectors"][self.my_sector]["path"]
        countries = sorted(tree.values(), key=lambda x: x["title"])
        self.my_group = None
        for country in countries:
            country["sectors"] = sectors = sorted(
                country["sectors"].values(), key=lambda x: x["title"]
            )

            for sector in sectors:
                sector["groups"] = groups = sorted(
                    sector["groups"].values(), key=lambda x: x["title"]
                )

                if sector["path"] == my_sector_path and sector["groups"]:
                    self.my_group = sector["groups"][0]["id"]
                for group in groups:
                    group["surveys"] = sorted(
                        group["surveys"].values(), key=lambda x: x["title"]
                    )

                sector["groups"] = [g for g in groups if g["surveys"]]
            country["sectors"] = [s for s in sectors if s["groups"]]
        countries = [c for c in countries if c["sectors"]]
        return countries

    def get_template_copy(self, container, template):
        """Get's the copy for the template

        The template is a survey that will be copied inside the newly
        created survey group.
        """
        copy = template._getCopy(container)

        # Reset the copied tree UIDs
        # We use a fake event to reuse the subscriber from plone.uuid that
        # resets the UIDs on copy
        fake_copy_event = ObjectCopiedEvent(None, None)
        addAttributeUUID(copy, fake_copy_event)
        for id, item in copy.ZopeFind(copy, search_sub=1):
            addAttributeUUID(item, fake_copy_event)

        # Set a time based id
        today = datetime.date.today()
        copy.id = today.isoformat()

        # Set the title to the copy
        title = self.request.locale.dates.getFormatter("date", length="long").format(
            datetime.date.today()
        )
        copy.title = title

        return copy

    def copyTemplate(self, source, target):
        target = self.context[target.id]  # Acquisition-wrap
        try:
            source._notifyOfCopyTo(target, op=0)
        except ConflictError:
            raise

        copy = self.get_template_copy(target, source)

        source_algorithm = aq_parent(source).evaluation_algorithm
        target_algorithm = self.request.form.get(
            "form.widgets.evaluation_algorithm", [source_algorithm]
        )
        if not isinstance(target_algorithm, str):
            target_algorithm = target_algorithm.pop()

        target.evaluation_algorithm = target_algorithm
        target._setObject(copy.id, copy)
        if source_algorithm != target_algorithm:
            from euphorie.content.risk import EnsureInterface
            from euphorie.content.risk import IRisk

            risks = [
                item
                for (id, item) in target.ZopeFind(target, search_sub=1)
                if IRisk.providedBy(item)
            ]
            for risk in risks:
                EnsureInterface(risk)

        if hasattr(copy, "published"):
            delattr(copy, "published")
        copy = target[copy.id]  # Acquisition-wrap
        copy.wl_clearLocks()
        copy._postCopy(target, op=0)

        wt = api.portal.get_tool("portal_workflow")
        if wt.getInfoFor(copy, "review_state") == "published":
            wt.doActionFor(copy, "retract")

        notify(ObjectClonedEvent(target[copy.id]))
        return copy

    def createAndAdd(self, data):
        obj = super().createAndAdd(data)
        obj = aq_inner(self.context)[obj.id]

        form = self.request.form
        if form["source"] != "scratch":
            sector = aq_inner(self.context)
            if form["source"] == "other":
                country_id = form["country"]
                country = aq_parent(aq_parent(sector))[country_id]
                bits = form["sector.%s" % country_id].split(".", 1)
                sector = country[bits[0]]
                surveygroup = sector[bits[1]]
                survey_id = form[f"survey.{country_id}.{surveygroup.id}"]
                survey = surveygroup[survey_id]
            else:
                sg_local = form["surveygroup.local"]
                surveygroup = sector[sg_local]
                survey = surveygroup[form["survey.local.%s" % sg_local]]

            survey = self.copyTemplate(survey, obj)
            self.immediate_view = survey.absolute_url()
        else:
            title = api.portal.translate(_("survey_default_title", default="Standard"))

            obj = aq_inner(self.context)[obj.id]
            survey = createContentInContainer(obj, "euphorie.survey", title=title)

            self.immediate_view = survey.absolute_url()
        return obj


class AddView(DefaultAddView):
    form = AddForm


class Unpublish(BrowserView):
    def unpublish(self):
        published_survey = self.context.published_survey

        wt = api.portal.get_tool("portal_workflow")
        if wt.getInfoFor(published_survey, "review_state") != "published":
            log.warning(
                "Trying to unpublish survey %s which is not marked as " "published",
                "/".join(published_survey.getPhysicalPath()),
            )
        else:
            wt.doActionFor(published_survey, "retract")
        notify(SurveyUnpublishEvent(published_survey))

    def post(self):
        action = self.request.form.get("action", "cancel")
        if action == "unpublish":
            self.unpublish()
            api.portal.show_message(
                _(
                    "message_unpublish_success",
                    default="This OiRA Tool is now no longer available in "
                    "the client.",
                ),
                self.request,
                "success",
            )
        else:
            api.portal.show_message(
                _("message_unpublish_cancel", default="Cancelled unpublish action."),
                self.request,
                "notice",
            )

        context = aq_inner(self.context)
        self.request.response.redirect(context.absolute_url())

    def __call__(self):
        if self.request.method == "POST":
            self.post()
        else:
            return super().__call__()


class VersionCommand(BrowserView):
    def __call__(self):
        surveygroup = aq_inner(self.context)
        action = self.request.form.get("action")
        survey_id = self.request.form.get("survey")
        response = self.request.response
        if action == "publish":
            response.redirect(f"{surveygroup.absolute_url()}/{survey_id}/@@publish")
        elif action == "unpublish":
            response.redirect("%s/@@unpublish" % surveygroup.absolute_url())
        elif action == "clone":
            response.redirect(
                "%s/++add++euphorie.survey?%s"
                % (surveygroup.absolute_url(), urlencode(dict(survey=survey_id)))
            )
        else:
            log.error("Invalid version command action: %r", action)



"""
Client
------

The view for the frontpage of the client site and some adapters to provide a
client user.
"""

from borg.localrole.interfaces import ILocalRoleProvider
from euphorie.client.interfaces import IClientSkinLayer
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.dexterity.content import Container
from plone.supermodel import model
from Products.membrane.interfaces.user import IMembraneUserObject
from zope.component import adapter
from zope.interface import directlyProvidedBy
from zope.interface import directlyProvides
from zope.interface import implementer
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserSkinType
from ZPublisher.BaseRequest import DefaultPublishTraverse


class IClient(model.Schema, IBasic):
    """The online client.

    The online client is implemented as a container with all available
    surveys. The default view for all survey elements inside this
    container is changed to the client user interface. This is done
    using a simple traversal adapter.
    """


@implementer(IClient)
class Client(Container):
    exclude_from_nav = True


@adapter(IClient)
@implementer(IMembraneUserObject)
class ClientUserProvider:
    """Expose the client as a user to the system.

    This is used as ownership for all client data.
    """

    def __init__(self, context):
        self.context = context

    def getUserId(self):
        return self.context.id

    def getUserName(self):
        return self.context.id


@adapter(IClient)
@implementer(ILocalRoleProvider)
class ClientLocalRolesProvider:
    """`borg.localrole` provider for :obj:`IClient` instances.

    This local role provider gives the client user itself the
    `CountryManager` local role. This allows publication of surveys
    inside the client since the publication machinery always runs under
    the client user.
    """

    def __init__(self, client):
        self.context = client

    def getRoles(self, principal_id):
        if principal_id == self.context.getId():
            return ("CountryManager",)
        return ()

    def getAllRoles(self):
        return [(self.context.getId(), ("CountryManager",))]


@adapter(IClient, Interface)
class ClientPublishTraverser(DefaultPublishTraverse):
    """Publish traverser to setup the skin layer.

    This traverser marks the request with IClientSkinLayer. We can not
    use BeforeTraverseEvent since in Zope 2 that is only fired for site
    objects.
    """

    skin_layer = IClientSkinLayer

    def publishTraverse(self, request, name):
        from euphorie.client.utils import setRequest

        setRequest(request)
        request.client = self.context  # XXX: remove me

        ifaces = [
            iface
            for iface in directlyProvidedBy(request)
            if not IBrowserSkinType.providedBy(iface)
        ]
        directlyProvides(request, self.skin_layer, ifaces)
        return super().publishTraverse(request, name)



from OFS.event import ObjectClonedEvent
from OFS.interfaces import IObjectClonedEvent
from plonetheme.nuplone.skin.interfaces import NuPloneSkin
from zope.interface import implementer
from zope.interface import Interface
from zope.interface.interfaces import IObjectEvent
from zope.interface.interfaces import ObjectEvent


class IEuphorieContentLayer(NuPloneSkin):
    """Marker interface for Euphorie content."""


class IObjectPublishedEvent(IObjectClonedEvent):
    """An object has been published by copying it to the client area."""


@implementer(IObjectPublishedEvent)
class ObjectPublishedEvent(ObjectClonedEvent):
    """An object has been published by copying it to the client area."""


class ISurveyUnpublishEvent(IObjectEvent):
    """A survey is being removed from the client."""


@implementer(ISurveyUnpublishEvent)
class SurveyUnpublishEvent(ObjectEvent):
    """A survey is being removed from the client."""


class IQuestionContainer(Interface):
    """Marker interface for objects that are used for grouping, but not for
    risks."""


class ICustomRisksModule(Interface):
    """Marker interface to mark a module specifically designated for adding
    custom risks."""



from euphorie.content.utils import getTermTitleByToken
from euphorie.content.utils import getTermTitleByValue
from euphorie.content.utils import parse_scaled_answers
from euphorie.content.utils import StripMarkup
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import unittest


class StripMarkupTests(unittest.TestCase):
    def StripMarkup(self, *a, **kw):
        return StripMarkup(*a, **kw)

    def testEmpty(self):
        obj = Mock()
        obj.description = ""
        self.assertEqual(self.StripMarkup(""), "")

    def testNoMarkup(self):
        self.assertEqual(self.StripMarkup("Test"), "Test")

    def testSingleTag(self):
        self.assertEqual(self.StripMarkup("Test <em>me</me>"), "Test me")

    def testMultipleTags(self):
        self.assertEqual(
            self.StripMarkup("Test <em>me</me> <strong>now</strong>"), "Test me now"
        )

    def testStrayBracket(self):
        self.assertEqual(self.StripMarkup("Test <em>me</em> >"), "Test me >")


class getTermTitleByValueTests(unittest.TestCase):
    def getTermTitleByValue(self, *a, **kw):
        return getTermTitleByValue(*a, **kw)

    def testUnknownValue(self):
        self.assertEqual(self.getTermTitleByValue(make_field([]), "dummy"), "dummy")

    def testKnownValue(self):
        field = make_field([(1, "token", "Title")])
        self.assertEqual(self.getTermTitleByValue(field, 1), "Title")


class TestGetTermTitleByToken(unittest.TestCase):
    def testUnknownToken(self):
        self.assertEqual(getTermTitleByToken(make_field([]), "dummy"), "dummy")

    def testKnownToken(self):
        field = make_field([(1, "token", "Title")])
        self.assertEqual(getTermTitleByToken(field, "token"), "Title")


class TestParseScaledAnswers(unittest.TestCase):
    def testEmpty(self):
        self.assertEqual(parse_scaled_answers(""), [])

    def testSimple(self):
        self.assertEqual(
            parse_scaled_answers("a\nb"),
            [{"text": "a", "value": "1"}, {"text": "b", "value": "2"}],
        )
        self.assertEqual(
            parse_scaled_answers("a\nb\nc\nd\ne\nf"),
            [
                {"text": "a", "value": "1"},
                {"text": "b", "value": "2"},
                {"text": "c", "value": "3"},
                {"text": "d", "value": "4"},
                {"text": "e", "value": "5"},
                {"text": "f", "value": "6"},
            ],
        )

    def testExtended(self):
        self.assertEqual(
            parse_scaled_answers("a|7\nb|3"),
            [{"text": "a", "value": "7"}, {"text": "b", "value": "3"}],
        )

    def testMixed(self):
        self.assertEqual(
            parse_scaled_answers("a|7\nb"),
            [{"text": "a", "value": "7"}, {"text": "b", "value": "2"}],
        )
        self.assertEqual(
            parse_scaled_answers("a\nb|42"),
            [{"text": "a", "value": "1"}, {"text": "b", "value": "42"}],
        )

    def testUgly(self):
        self.assertEqual(
            parse_scaled_answers("\n\n hello world |  7\n\n\n  b  \n\n\nc|\n\n\n"),
            [
                {"text": "hello world", "value": "7"},
                {"text": "b", "value": "2"},
                {"text": "c", "value": "3"},
            ],
        )
        # If you really want, an answer can contain a literal pipe.
        self.assertEqual(
            parse_scaled_answers("a|b|7"),
            [{"text": "a|b", "value": "7"}],
        )
        self.assertEqual(
            parse_scaled_answers("a||7"),
            [{"text": "a|", "value": "7"}],
        )


class Mock:
    pass


def make_field(terms):
    terms = [SimpleTerm(value, token, title=title) for (value, token, title) in terms]
    field = Mock()
    field.vocabulary = SimpleVocabulary(terms)
    return field
